package CLASSES::Build;
use strict;
use CLASSES::DBClass;
use CLASSES::Chromosome;
use QUERY::query_factory;
use TOOLS::assert;

our @VERSION = '1.0';
our @ISA = qw(CLASSES::DBClass);

################################################
# Required methods
sub _init {
    return $_[0];
}

################################################
# Other methods

#####################
# Name: get_chromosome
# Purpose: Dynamic method to get chromosome i of the build
# Parameters: i
# Output: The chromosome object (or the list of all chromosomes if i undefined)
sub get_chromosome {
     my $self = assert_dynamic(shift);
     assert(@_ <= 1, "Bad parameters\n");

     if (@_==1) {
	 my $i = shift;

	 return scalar CLASSES::Chromosome->get("dbh"=>$self->get_dbh(), "where"=>"Chromosome.number=?gc_num? and Chromosome.build_id=?gc_build?", "values"=>{"gc_num"=>$i, "gc_build"=>$self->get_id()}, "saved_query"=>"build_get_chr");
     }
     else {
	 return CLASSES::Chromosome->get("dbh"=>$self->get_dbh(), "where"=>"Chromosome.build_id=?gc_build?", "values"=>{"gc_build"=>$self->get_id()}, "command"=>"order by number", "saved_query"=>"build_get_all_chr");
     }
}

#######################
# Name: get_by_number
# Purpose: Get the indicated build
# Parameters: database handle, build number
# Output: The build object
sub get_by_number {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $i = shift || die("CLASSES::Build::get_by_build: Must give a build number\n");

    return scalar $class->get("dbh"=>$dbh, "where"=>"build_number=?gb_bn?", "values"=>{"gb_bn"=>$i}, "saved_query"=>"get_by_build");
}


#######################
# Name: current_build_number
# Purpose: (static) Return the latest build number
# Parameters: dbh
# Output: The latest build number
sub current_build_number {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $query = QUERY::Cache->get($dbh, $class, "current_build");
    unless ($query) {
	$query = create_query("type"=>"SVSelect",
			      "dbh"=>$dbh,
			      "table"=>"Build",
			      "return_field"=>"build_number",
			      "function_fields"=>{"Build.build_number"=>"max"});
	QUERY::Cache->set($dbh, $class, "current_build", $query);
    }
    $query->execute();
    return ($query->fetch_values())[0];
}

######################
# Name: current_build
# Purpose: (static) Return the latest build number
# Parameters: dbh
# Output: The lastest build object
sub current_build {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);

    my $id = CLASSES::Build->current_build_number($dbh);
    return CLASSES::Build->get_by_number($dbh, $id);
}

######################
# Name: chr_numbers
# Purpose: A (sorted) list of all chromosome numbers for the build
# Parameters: None
# Output: A sorted list of strings
sub chromosome_numbers {
    my $self = assert_dynamic(shift);
    my @chr_array = CLASSES::Chromosome->get("dbh"=>$self->get_dbh(),
					     "where"=>"build_id = ".$self->get_id());
    my @number_array = map {$_->get_number()} @chr_array;

    return sort {if ($a eq 'M') {return 1};  if ($b eq 'M') {return -1};  if ($a eq 'M' || $a eq 'X' || $a eq 'Y') {return $a cmp $b}; if ($b eq 'M' || $b eq 'X' || $b eq 'Y') {return $a cmp $b}; $a <=> $b} @number_array;
}


1;
