package CLASSES::Chromosome;
use strict;
use CLASSES::DBClass;
use CLASSES::Build;
use TOOLS::assert;
use TOOLS::genome;

our @VERSION = '1.0';
our @ISA = qw(CLASSES::DBClass);

#-------------------------------------------------------------------------
# Required Functions
sub _init {
    my $self = shift;
    my $dbh = $self->get_dbh();
    my $genome_label = $self->get_build_id()."-".$self->get_number();
    my $genome = DB::Cache->get($dbh, "genome", $genome_label);
    unless ($genome) {
	my $build = CLASSES::Build->get_pk($dbh, $self->{"_build_id"});
	my $file = $build->get_directory().$self->{"_file_name"};
	$genome = genome->new($file);
        DB::Cache->set($dbh, "genome", $genome_label, $genome);
    }

    $self->correct_attribute("size", $genome->get_size());

    return $self;
}

#-------------------------------------------------------------------------
# Other functions

#####################
# Name: get_chromosome
# Purpose: (static) Get the chromosome object from a build number and chromsome number
# Parameters: A database handle, a build number and a chromosome number
# Output: A chromosome object
sub get_chromosome {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my ($build, $chr) = @_;

    return scalar CLASSES::Chromosome->get("dbh"=>$dbh, "join_tables"=>["Build"], "where"=>"Build.build_number=?gc_build_num? and Build.id=Chromosome.build_id and Chromosome.number=?gc_chr_num?", "values"=>{"gc_build_num"=>$build, "gc_chr_num"=>$chr}, "saved_query"=>"get_chromosome");
}

#####################
# Name: get_genome
# Purpose: (dynamic) get the associated genome object
# Parameters: None
# Output: The genome object
sub get_genome {
    my $self = assert_dynamic(shift);
    my $genome_label = $self->get_build_id()."-".$self->get_number();
    return DB::Cache->get($self->get_dbh(), "genome", $genome_label);
}

#####################
# Name: get_build
# Purpose: (dynamic) get the associated build object
# Parameters: None
# Output: A build object
sub get_build {
    my $self = assert_dynamic(shift);
    return CLASSES::Build->get_pk($self->get_dbh(), $self->get_build_id());
}

#####################
# Name: get_segment_by_position
# Purpose: (dynamic) Get a segment by position number, or between two positions
# Parameters: The position number of the segment (or none)
# Output: The segment object, or the list of all segments (ordered by position number)
sub get_segment_by_position {
    my $self = assert_dynamic(shift);
    assert(@_ <= 1, "Bad parameters\n");

    if (@_==1) {
	my $i = shift;
	return scalar CLASSES::Chromosome_Segment->get("dbh"=>$self->get_dbh(), "where"=>"chr_id=?gsbp_chr_id? and position=?gsbp_position?", "values"=>{"gsbp_chr_id"=>$self->get_id(), "gsbp_position"=>$i}, "saved_query"=>"get_segment_by_position");
    }
    else {
	return CLASSES::Chromosome_Segment->get("dbh"=>$self->get_dbh(), "where"=>"chr_id=?gsbp_chr_id?", "values"=>{"gsbp_chr_id"=>$self->get_id(), "command"=>"order by position", "saved_query"=>"get_all_segments"});
    }
}

#####################
# Name: get_segments
# Purpose: (dynamic) Get all segments intersecting the given interval on the chromosome
# Parameters: start coord, end coord
# Output: A list of segments
sub get_segments {
    my $self = assert_dynamic(shift);
    my $start = shift;
    my $finish = shift || $start;

    assert($start, "Must provide at least one coordinate\n");
    assert($start >= 1 && $start <= $finish && $finish <= $self->get_size(), "Bad coordinates: $start, $finish\n");

    my $left_where = QUERY::Where->new("?gs_start? between start and finish");
    my $right_where = QUERY::Where->new("?gs_finish? between start and finish");
    my $interior_where = QUERY::Where->new("(start between ?gs_start? and ?gs_finish?) and (finish between ?gs_start? and ?gs_finish?)");
    my $where = QUERY::Where->new("binary_op"=>"or", "clauses"=>[$left_where, $right_where, $interior_where]);

    my @arr = CLASSES::Chromosome_Segment->get("dbh"=>$self->get_dbh(), "where"=>[$where, "chr_id=?gs_chr_id?"], "command"=>"order by position", "values"=>{"gs_start"=>$start, "gs_finish"=>$finish, "gs_chr_id"=>$self->get_id()}, "saved_query"=>"get_segments");

    return @arr;
}

####################
# Name: get_segment_offset
# Purpose: (dynamic) Get the segment and offset into the segment of a given point
# Parameters: A single coordinate on the chromosome
# Output: A segment object and the offset into the segment
sub get_segment_offset {
    my $self = assert_dynamic(shift);
    my $coord = shift;
    assert($coord && $coord >= 1 && $coord <= $self->get_size(), "Coordinate out of range");

    my ($seg) = $self->get_segments($coord);
    return ($seg, $seg->absolute_to_offset($coord));
}

######################
# Name: get_sequence
# Purpose: (dynamic) Print the sequence from the specified coords
# Parameters: Start coord (relative to + strand), finish coord, strand (+/- default +)
# Output: A string
sub get_sequence {
    my $self = assert_dynamic(shift);
    my ($start, $finish, $strand) = @_;
    assert($start >= 1 && $start <= $finish && $finish <= $self->get_size(), "Bad coordinates given");

    my $g = $self->get_genome();
    my $len = $finish - $start + 1;

    return (defined $strand && $strand eq '-') ?
	$g->get_reverse_complement($start, $len) :
	$g->get_sequence($start, $len);
}

1;
