package CLASSES::Collection;
use strict;

use CLASSES::DBClass_Versioned_Free;
use TOOLS::assert;
use QUERY::query_factory;
use TaxID::tax_id;

our @VERSION = '1.0';
our @ISA = qw(CLASSES::DBClass_Versioned_Free);

#-------------------------------------------------------------------------
# Required Functions
sub _vinit {
    my $self = shift;
    my %params = @_;

    my $contents = $params{contents} || $params{content};
    if ($contents) {
	assert(ref($contents) eq "ARRAY", "Not an array");
	$self->_set_contents(@$contents);
    }

    return $self->_cinit(@_);;
}

#-------------------------------------------------------------------------
# Extra functions

#################
# Name: mapping_table
# Purpose: Return the name of the table mapping the collection to the object
sub collection_type() {
    assert(0==1, "Virtual function; must be implmeneted for concrete class");
}

sub mapping_table() {
    return $_[0]->collection_type()."_to_Collection";
}


sub column() {
    return "_".$_[0]->collection_type()."_ID";
}

sub last() {
    return "_last_".$_[0]->collection_type();
}

sub content_compare() {
    return ($_[1]->get_id() <=> $_[2]->get_id());
}

######################
# Name: new
# Purpose: Assure new is not called as an (abstract) Collection object
# Input: Whetever required by new
# Output: The new object
sub new {
    my $class = assert_abstract(shift);

    return $class->CLASSES::DBClass_Versioned_Free::new(@_);
}

######################
# Name: _set_content
# Purpose: (dynamic) Set the content of the object
# Parameters: None
# Output: None
sub _set_contents {
    my $self = assert_dynamic(shift);
    my @contents = @_;

    $self->{_contents} = \@contents;
}

######################
# Name: get
# Purpose: Augment the DBClass_Versioned_Free::get method to extract the contents and extension information
# Parameters: No additional parameters
# Output: The array of objects
sub get {
    my $class = assert_static(shift);
    my @objs = $class->CLASSES::DBClass_Versioned_Free::get(@_);

    # Leave the contents field blank for lazy evaluation
    assert(wantarray || @objs<=1, "Can't call in scalar context when number of results is more than 1\n");

    return wantarray ? @objs : shift @objs;
}

######################
# Name: get_content
# Purpose: (Dynamic) Return the list of contents in the collection
# Parameters: interger i (optional)
# Output: ith content or an array of all contentss
sub get_contents {
    my $self = assert_dynamic(shift);
    my $i = shift;
    my $type = $self->collection_type();
    my $map = $self->mapping_table();
    my $column = $self->column();
    my $dbh = $self->get_dbh();

    $self->{_contents} = ["CLASSES::$type"->get(dbh => $dbh,
					       join_tables => [$self->mapping_table()],
					       where => ["$map.$column = $type.id",
							 "$map._Collection_ID = ?col_id?"],
					       command => "order by $map._sequence_position",
					       values => {col_id => $self->get_id()},
					       saved_query => "get_collection_$type")]
		   unless $self->{_contents};

    return $i ? $self->{_contents}->[$i] : @{$self->{_contents}};
}

#####################
# Name: _save_contents
# Purpose: (dynamic) Save the contents in the content_to_collection tables.  (Assumes this is a new collection.)
# Parameters: None
# Output: None
sub _save_contents {
    my $self = assert_dynamic(shift);
    my $class = ref($self);
    my $dbh = $self->get_dbh();
    my $map = $self->mapping_table();
    my $type = $self->collection_type();
    my $column = $self->column();
    my $last = $self->last();

    my $query_name = "save_contents";
    my $query = QUERY::Cache->get($dbh, $class, $query_name);
    unless ($query) {
	$query = create_query(type => "Insert",
			      dbh => $dbh,
			      table => $map,
			      attributes => [$column, qw(_Collection_ID _sequence_position), $last]);
	QUERY::Cache->set($dbh, $class, $query_name, $query);
    }

    my @contents = sort {$self->content_compare($a, $b)} $self->get_contents();
    for (my $i=0; $i < @contents; $i++) {
	my $F = $contents[$i];
	$query->execute($column=>$F->get_id(), _Collection_ID=>$self->get_id(), _sequence_position=>$i, $last=>($i==@contents-1 ? 'T' : 'F'));
    }

}

#####################
# Name: insert
# Purpose: Augment the DBClass_Versioned_Free::insert function to insert the content list to the database
# Parameters: None
# Ouput: None
sub insert {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my @objs = ref($proto) ? ($proto) : @_;

    $class->CLASSES::DBClass_Versioned_Free::insert(@objs);
    for (@objs) {
	$_->_save_contents();
    }
}

##########################
# Name: size
# Purpose: (Dynamic) Number of contents in the collection
# Paramaters: None
# Output: Integer
sub size {
    my $self = assert_dynamic(shift);
    my @contents = $self->get_contents;
    return scalar @contents;
}

######################
# Name: class_copy
# Purpose: Static method to copy the contents of the second object into the first object
# Parameter: Operand object, operator object
# Output: None
sub class_copy {
    my $class = assert_static(shift);
    my ($lvalue, $rvalue) = (shift, shift);
    $lvalue->{_contents} = $rvalue->{_contents};

    $class->CLASSES::DBClass_Versioned_Free::class_copy($lvalue, $rvalue);    
}

######################
# Name: get_by_content
# Purpose: (static) Return all sets containing the given_content
# Parameter dbh object, content object, search parameter hash
# Output: None
sub get_by_contents {
    my $class = assert_static(shift);
    my $type = $class->collection_type();
    my $map = $class->mapping_table();
    my $column = $class->column();
    my $name = $class->class_name();
    my $dbh = assert_dbh(shift);
    my $F = shift;  assert(ref($F) && ref($F) eq "CLASSES::$type", "Non-$type passed as parameter");
    my %params = @_;


    return $class->get(dbh=>$dbh,
		       join_tables => [$map],
		       where => ["$map.$column = ?object_id?",
				 "$map._Collection_ID=$name.id"],
		       values => {object_id => $F->get_id()},
		       saved_query => "get_by_$type");
}


1;

