package CLASSES::DBClass;
our $VERSION = '1.0';
use strict;

use TOOLS::get_date_string;
use TOOLS::assert;
use TOOLS::constants;
use QUERY::Cache;
use QUERY::query_factory;
use DB::Cache;

#######################
# Name: new
# Purpose: Base-class constructor / copy constructor (abstract)
# Parameters: (hash style -- none if used as copy constructor)
#             dbh => Database handle
#             in_synch => is synchronized with db (default 0)
#             in_db => is in the db (default 0)
#             <attribute> => <value> (for all class attributes)
# Output: The new object
# Note: A copied object will have an undefined id
sub new {
    my $proto = assert_abstract(shift);

    if (ref($proto)) {
	# Copy constructor
	my $self = $proto;
	my $class = ref($self);
	my $new_obj = $class->new("dbh"=>$self->get_dbh());
	$class->class_copy($new_obj, $self);
	return $new_obj;
    }

    # Creation Constructor
    my $class = $proto;
    my %params = @_;
    my $self = bless {}, $class;

    for ($class->get_attribute_list()) {
	next if $_ eq "id";
	my $att = "_$_";
	my $get_fun = "$class\:\:get_$_";
	my $set_fun = "$class\:\:set_$_";
	{
	    no strict "refs";
	    unless (defined *{$get_fun}) {
		*{$get_fun} = sub {my $self = assert_dynamic(shift); return $self->{$att}};
	    }
	    unless (defined *{$set_fun}) {
		*{$set_fun} = sub {my $self = assert_dynamic(shift); $self->{$att} = $_[0]; $self->set_synch(undef); $self->set_id(undef)};
	    }
	}

	$self->{$att} = $params{$_};
    }

    $self->set_id($params{"id"});
    $self->{"_dbh"} = assert_dbh($params{"dbh"});
    $self->_init(%params);

    $self->set_synch($params{"synch"});
    $self->set_in_db($params{"in_db"});
 
    return $self;
}

#######################
# Name: DESTROY
# Purpose: None
# Parameters: None
# Output: None
sub DESTROY {
}

#######################
# Name: class_name
# Purpose: Get the name of the class (without the CLASSES:: prefix)
# Parameters: None
# Output: The string
sub class_name {
    my $class = assert_static(shift);
    my $name = ($class =~ /::([^:]+)$/g)[0];
    my $fun = "$class\:\:class_name";
    {
	no strict "refs";
	*{$fun} = sub {assert_static(shift, $fun); return $name};
    }
    return $name;
}

#######################
# Name: get_attribute_hash
# Purpose: Return the attribute hash (creating it in the first call)
# Parameters: None
# Referene to the attribute hash
sub get_attribute_hash {
    my $class = assert_static(shift);
    my $var = "$class\:\:attribute_hash";
    my $file = $class->class_name().".cl";
    my $fun = "$class\:\:get_attribute_hash";

    my $hash_ref = $class->_read_attributes($file);
    {
	no strict "refs";
	*{$fun} = sub {assert_static(shift, $fun); return $hash_ref};
    }
    return $hash_ref;
}

#######################
# Name: _init
# Purpose: Virtual function for augmenting constructor
# Parameters: None
# Output: $self
sub _init {
    die("DBClass::_init: Virtual function\n");
}

#######################
# Name: get_attribute_hash
# Purpose: A virtual static function returning the attribute hash reference
# Parameters: None
# Output: Reference to hash of attributes (to types)
# sub get_attribute_hash {
#     die("DBQuery::get_atribute_list: Virtual function\n");
# }

#######################
# Name: get_attribute_list
# Purpose: A static function returning a list of all attributes
# Parameters: None
# Output: List of all attributes
sub get_attribute_list {
    my $class = assert_static(shift);
    return keys %{$class->get_attribute_hash()}
}

#######################
# Name: is_attribute
# Purpose: A static function to determin if a given string is an attribute of the class
# Parameters: A string
# Output: boolean
sub is_attribute {
    my $class = assert_static(shift);
    my $str = shift;    
    return exists($class->get_attribute_hash()->{$str});
}

#######################
# Name: get_dbi
# Purpose: dynamic method to get an object's database handle
# Parameters: None
# Output: The dbi handle
sub get_dbh {
    my $class = assert_dynamic(shift);
    return $class->{"_dbh"};
}

#######################
# Name: _read_attributes
# Purposes: Read a .cl file an return a hash of the attribute/type mapping
# Parameters: A .cl file
# Output: A hash reference
sub _read_attributes {
    my $self = assert_static(shift);
    my $file = shift;
    assert($file =~ /\.cl$/, "Bad file name: $file\n");
    
    open(IFILE, TOOLS::constants::DESCRIPTION_FILE.$file) or die((caller())[3].": File does not exist: $file\n");
    my %H;
    while (my $line = <IFILE>) {
	my @arr = split("\t+", $line);
	$H{$arr[1]} = $arr[2] if $arr[0] eq "att";
    }
    close(IFILE);
    
    return \%H;
}

#####################
# Name: get_attribute
# Purpose: Get the value of a given attribute
# Parameters: An attribute string
# Output: The value of the attribute
sub get_attribute {
    my $self = assert_dynamic(shift);
    my $str = shift;
    
    assert(ref($self)->is_attribute($str), "Bad attribute for ".ref($self).": $str\n");
    my $fun = "get_$str";
    return $self->$fun();
}

#####################
# Name: set_attribute
# Purpose: Set the value of a given attribute
# Parameters: An attribute string, a value
# Output: None
sub set_attribute {
    my $self = assert_dynamic(shift);
    my ($str, $value) = @_;
    assert(ref($self)->is_attribute($str), "Bad attribute: $str\n");

    my $fun = "set_$str";
    $self->$fun($value);
}

########################
# Name: get_attribute_type
# Purpose: A method (static or dynamic) to get the type of the given attribute
# Parameters: A string
# Ouput: A string
sub get_attribute_type {
    my $class = ref($_[0]) || $_[0];
    my $str = $_[1];
    assert($class->is_attribute($str), "get_attribute_type: Bad attribute: $str");
    return $class->get_attribute_hash()->{$str};
}

#######################
# Name: insert
# Purpose: Static or dynamic
#          Static: Add the given objects into the table
#          Dynamic: Add the single object in to the table
# Parameters:
#          Static: One or more objects, all using the same database handle
#          Dynamic: None
# Output: None
# Side effects: Will assign objects id numbers and update id attribute
# Note: Cannot insert an object that has an id number
sub insert {
    my $proto = shift;
    my ($class, $static) = ref($proto) ? (ref($proto), 0) : ($proto, 1);

    assert(@_==0, "Dynmic version takes no parameters") unless $static;
 
    my @objs = $static ? @_ : ($proto);
    my $dbh = $objs[0]->get_dbh();
    my $query = QUERY::Cache->get($dbh, $class, "insert");
    unless (defined $query) {
	my $table = $class->class_name();
	my @attributes = $class->get_attribute_list();
	$query = create_query("type"=>"Insert", 
			      "dbh"=>$dbh, 
			      "table"=>$table, 
			      "attributes"=>\@attributes);
	QUERY::Cache->set($dbh, $class, "insert", $query);
    }
    for my $obj (@objs) {
	assert($obj->get_dbh() == $dbh, "Differing database handle: $dbh\n");
	assert(!defined($obj->get_id()), "Can't insert objects with id numbers\n");
	$obj->correct_attribute("entry_date", get_date_string::get_date_string());
	$query->execute($obj->all_values());
	$obj->set_attribute("id", DB::db::get_pk_value($dbh));
	$obj->set_in_db(1);
	$obj->set_synch(1);
	DB::Cache->set($dbh, $class, $obj->get_id(), $obj);
    }
}

#######################
# Name: remove
# Purpose: Dynamic method to remove the object from the database
# Parameters: None
# Output: None
sub remove {
    my $self = assert_dynamic(shift);
    my $class = ref($self);
    my $where;

    assert($self->is_in_db(), "Removing $class object not in database\n");
    my $dbh = $self->get_dbh();
    my $query = QUERY::Cache->get($dbh, $class, "remove");
    unless ($query) {
	my $table = $class->class_name();
	$query = create_query("type"=>"Remove", 
			      "dbh"=>$dbh, 
			      "table"=>$table, 
			      "where"=>"id=?remove_id?");
	QUERY::Cache->set($dbh, $class, "remove", $query);
    }
    $query->execute("remove_id"=>$self->get_id());
    $self->set_synch(undef);
    $self->set_in_db(undef);
    DB::Cache->remove($dbh, $class, $self->get_id());
    $self->set_attribute("id", undef);
}

#######################
# Name: update
# Purpose: Dynamic method to update the class in the table
# Parameters: None
# Output: None
sub update {
    my $self = assert_dynamic(shift);
    my $class = ref($self); 

    assert($self->is_in_db(), "Updating $class object not in database");
    assert($self->get_id(), "Can't update an object with an id number");
    return if ($self->in_synch());

    my $dbh = $self->get_dbh();
    my $query = QUERY::Cache->get($dbh, $class, "update");
    unless ($query) {
	my $table = ref($self)->class_name();
	my @attributes = $class->get_attribute_list();
	$query = create_query("type"=>"Update",
			      "dbh"=>$self->get_dbh(), 
			      "table"=>$table, 
			      "attributes"=>\@attributes, 
			      "where"=>"id=?new_id?");
	QUERY::Cache->set($dbh, $class, "update", $query);
    }

    $query->execute($self->all_values(), "new_id"=>$self->get_id());
    $self->set_synch(1);
}
    


#######################
# Name: get
# Purpose: A static method to extract rows from a database
# Parameters (hash style)
#            dhb => Dataebase handle
#            functional_fields => A hash of attribute names to funtions to be imposed on them
#            join_tables => reference to an array of additional tables to be used in a join
#            where => where object
#            values => hash of where_values for the where object
#            saved_query => name of saved query
#            where_filter => where string for additional filtering
#            get_components => Instead if preforming the query, return an array containing the join tables array (reference) and where object
# Output: A list of new objects conforming to the where statement (static version)
sub get {
    my $class = assert_static(shift);
    my %params = @_;

    my $dbh = assert_dbh($params{"dbh"});
    my $join_tables_ref = $params{"join_tables"};

    if ($params{get_components}) {
	assert(!$params{where_filter}, "Cannot use where_filter parameter with get_components parameter");
	my %tables;
	@tables{@$join_tables_ref} = undef if $join_tables_ref && @$join_tables_ref;
	my $where;
	if (ref($params{where}) eq 'QUERY::Where') {
	    $where = $params{where};
	}
	elsif (!ref($params{where})) {
	    $where = QUERY::Where->new(clauses => [$params{where}]);
	}
	else {
	    $where = QUERY::Where->new(clauses => $params{where});
	}
	my @arr = ([keys %tables], $where, $params{values} || {});
	return \@arr;
    }

    my ($saved, $query) = (undef, undef);

    unless ($params{where_filter}) {
	$saved = $params{"saved_query"};
	$query = $saved ? QUERY::Cache->get($dbh, $class, $saved) : undef;
    }

    unless ($query) {
	my $table_name = $class->class_name();
	my @fields = $class->get_attribute_list();

	my %H;

	my %table_hash;
	$table_hash{$table_name} = \@fields;

	@table_hash{@$join_tables_ref} = undef if $join_tables_ref && @$join_tables_ref;
	$query = create_query("type"=>"Select",
			      "dbh"=>$dbh, 
			      "table_hash"=>\%table_hash,
			      "functional_fields"=>$params{"funtional_fields"},
			      "where"=> $params{where_filter} ? [$params{where}, $params{where_filter}] : $params{where}, 
			      "command"=>$params{"command"}) unless defined $query;
	QUERY::Cache->set($dbh, $class, $saved, $query) unless !$saved;
    }

    my %values = defined($params{"values"}) ? %{$params{"values"}} : ();
 
    if ($params{"return_query"}) {
       assert(!$params{"values"}, "Can't pass values when returning the query");
       assert(!wantarray, "Can't call in array context when returning the query");
       return $query;
    }

    $query->execute(%values, "debug-explain"=>$params{"debug-explain"}, "debug-print"=>$params{"debug-print"});

    my @objs;
    for ($query->fetch_all()) {
	my $obj = $class->new("dbh"=>$dbh, "synch"=>1, "in_db"=>1, %$_);
	my $old = DB::Cache->get($dbh, $class, $obj->get_id());
	if ($old) {
	    unless ($old->in_synch()) {
		$class->class_copy_exact($old, $obj);
	    }
	    $obj = $old;
	}
	else {
	    DB::Cache->set($dbh, $class, $obj->get_id(), $obj);
	}
	push(@objs, $obj);
    }

    assert(wantarray || @objs<=1, "Can't call in scalar context when number of results is more than 1\n");

    return wantarray ? @objs : shift @objs;
}

#######################
# Name: get_pk
# Purpose: A static method to get objects by their primary keys
# Parameters: A database handle and one primary key (and extra paramters for internal calls)
# Output: The object
sub get_pk {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $primary_key = shift;
    assert($primary_key, "Must provide a primary key\n");

    my %extra_params = @_;
    $extra_params{"saved_query"} = "saved_query" unless $extra_params{"saved_query"};
    return $class->get("dbh"=>$dbh, "where"=>"id=?get_pk_id?", "values"=>{"get_pk_id"=>$primary_key}, %extra_params);
}

    

#######################
# Name: set_id
# Purpose: (dynamic) Set the id of the ojbect
# Parameters: The new value
# Output: None
sub set_id {
    my $self = assert_dynamic(shift);
    $self->{"_id"} = shift;
    $self->set_synch(undef);
}

#######################
# Name: get_id
# Purpose: (dynamic) Get the id of the object
# Parameters: The new value
# Output: None
sub get_id {
    my $self = assert_dynamic(shift);
    return $self->{"_id"};
}



#######################
# Name: in_synch / set_synch / is_in_db / set_in_db
# Purpose: Dynamic methods to check/set the synch/in_db status
# Parameters: None or bool
# Output: bool / none
sub in_synch {
    my $self = assert_dynamic($_[0]);
    return $self->{"_synch"} && $self->is_in_db();
}

sub set_synch {
    my $self = assert_dynamic($_[0]);
    $self->{"_synch"} = $_[1];
}

sub is_in_db {
    my $self = assert_dynamic($_[0]);
    return $self->{"_in_db"};
}

sub set_in_db {
    my $self = assert_dynamic($_[0]);
    $self->{"_in_db"} = $_[1];
}


######################
# Name: class_copy
# Purpose: Static method to copy the contents of the second object into the first object
# Parameter: Operand object, operator object
# Output: None
# Note: lValue is not assigned an id (or version number)
sub class_copy {
    my $class = assert_static(shift);
    my ($lvalue, $rvalue) = @_;
    assert(ref($lvalue) eq ref($rvalue) && ref($lvalue) eq $class, "Bad objects\n");
    assert($lvalue->get_dbh() == $rvalue->get_dbh(), "Object not in same database\n");

    for ($class->get_attribute_list()) {
	$lvalue->correct_attribute($_, $rvalue->get_attribute($_));
    }
    $lvalue->reset_id();
    $lvalue->set_in_db(0);
    $lvalue->set_synch(0);
}

######################
# Name: class_copy_exact
# Purpose: (static) Class copy, preserving
# Parameter: Operand object, operator object
# Output: None
sub class_copy_exact {
    my $class = assert_static(shift);
    my ($lvalue, $rvalue) = @_;

    $class->class_copy($lvalue, $rvalue);
    $lvalue->set_id($rvalue->get_id());
}

####################
# Name: all_values
# Purpose: Return a hash of all attributes to their values
# Parameters: None
# Output: Hash
sub all_values {
    my $self = assert_dynamic(shift);
    my %H;
    for (ref($self)->get_attribute_list()) {
	$H{$_} = $self->get_attribute($_);
    }
    return %H;
}

######################
# Name: correct_attribute
# Purpose: (dynamic) Correct the value of an attibute (without resetting the id)
# Parameters: attribute name, new value
# Output: None
sub correct_attribute {
    my $self = assert_dynamic(shift);
    my ($att, $value) = @_;
    my $id = $self->get_id();
    my $fun = "set_$att";
    $self->$fun($value);
    $self->set_id($id);
}

#######################
# Name: get_attribute_array 
# Purpose: (dynamic) Print a string of all values
# Paramters: list of attributes to print (default = all)
# Output: A formated string
sub get_attribute_array {
   my $self = assert_dynamic(shift);
   my @atts = @_>0 ? @_ : sort {return -1 if $a eq "id" || $b eq "entry_date"; return 1 if $b eq "id" || $a eq "entry_date"; return -1 if $a le $b; return 1 if $b le $a; return 0} ref($self)->get_attribute_list();
   return map {$self->get_attribute($_) || "."} @atts;
}

######################
# Name: reset_id
# Purpose: (dynamic) Reset the objects id to udef
# Parameters: None
# Output: None
sub reset_id {
    my $self = assert_dynamic(shift);
    $self->set_id(undef);
}

######################
# Name: get_by_entry_date
# Purpose: (state) Get all objects with an entry date before/between the given entry dates
# Parameters: database handle, single date (for < then) or two dates (inclusive)
# Output: All objects of the class falling within the date constraint
sub get_by_entry_date {
    my $class = assert_abstract(shift);
    my $dbh = assert_dbh(shift);
    my $date1 = shift;
    my $date2 = shift;

    my $query;
    if ($date2) {
	$query = QUERY::Cache->get($dbh, $class, "double-date");
	unless ($query) {
	    $query = create_query(type => "SVSelect",
				  dbh => $dbh,
				  table => $class->class_name(),
				  return_field => "id",
				  all_versions => 1,
				  where => "entry_date > ?d1? and entry_date < ?d2?");
	    QUERY::Cache->set($dbh, $class, "double", $query);
	}
	$query->execute(d1 => $date1, d2 => $date2);
    }
    else {
	$query = QUERY::Cache->get($dbh, $class, "double-date");
	unless ($query) {
	    $query = create_query(type => "SVSelect",
				  dbh => $dbh,
				  table => $class->class_name(),
				  return_field => "id",
				  all_versions => 1,
				  where => "entry_date < ?d1?");
	    QUERY::Cache->set($dbh, $class, "double", $query);
	}
	$query->execute(d1 => $date1);
    }

    return map {CLASSES::Pseudogene->get_pk($dbh, $_)} $query->fetch_values();
}

#####################
# Name: _modify_object_in_db
# Purpse: (dynamic) Allow for a modification of attributes in the object and db without other side-effects
# Parameters: Attribute name, new value
# Output: None
# Comment: NOT FOR USE BY USER!!!
sub _modify_object_in_db {
    my $self = assert_dynamic(shift);
    my $class = ref($self);
    my $dbh = $self->get_dbh();

    my ($att, $value) = @_;

    my $query_name = "_modify_$att\_in_db";
    my $query = QUERY::Cache->get($dbh, $class, $query_name);
    unless ($query) {
	$query = create_query("type"=>"Update",
			      "dbh"=>$dbh,
			      "table"=>$class->class_name(),
			      "attributes"=>["$att"],
			      "where"=>"id = ?id?");
	QUERY::Cache->set($dbh, $class, $query_name, $query);
    }
    $query->execute("$att"=>$value,"id"=>$self->get_id());
    $self->{"_$att"} = $value;
}

1;
 
