package CLASSES::DBClass_Versioned;
use strict;
use CLASSES::DBClass;
use CLASSES::Build;   # Needed to compute current build number
use QUERY::query_factory;
use QUERY::Where;
use TOOLS::assert;
use Data::Dumper;

our @VERSION = '1.0';
our @ISA = qw(CLASSES::DBClass);

########################
# Name: new
# Purpose: Perform some static modification to the class before calling the DBClass::new consturctor
# Input: Same as DBClass::new
# Outuput: None
# Note: Only needed once per class -- hence bypassed after first call
sub new {
    my $class = assert_abstract(assert_static(shift));
    my @params = @_;


    for ($class->get_attribute_list()) {
	next if $_ eq "id" or $_ eq "version";
	my $att = "_$_";
	my $set_fun = "$class\:\:set_$_";
	{
	    no strict "refs";
	    unless (defined *{$set_fun}) {
		*{$set_fun} = sub {my $self = assert_dynamic(shift); $self->{$att} = $_[0]; $self->set_synch(undef); $self->set_id(undef); $self->set_version(undef)};
	    }
	}
    }

    ############# Set up the bypass
    {
	my $fun = "$class\:\:new";
	no strict "refs";
	no warnings;
	*{$fun} = \&CLASSES::DBClass::new;
    }
    #############

    return $class->CLASSES::DBClass::new(@params);
}
	

########################
# Name: _init
# Purpose: Stuff to do after the DBClass consturctor
# Parameters: None
# Output: The object
sub _init {
    my $self = shift;
    assert(ref($self) ne "CLASSES::DBClass_Versioned", "DBClass_Versioned is abstract\n");
    return $self->_vinit(@_);
}

########################
# Name: get_version_table
# Purpose: Static method to get the name of the version table
# Paramters: None
# Output: String
sub get_version_table {
    my $class = assert_static(shift);

    my $l = length("CLASSES\:\:");
    return substr($class, $l, length($class)-$l)."_Version";
}

########################
# Name _current_version_numbers
# Purpose: Static method to get the current version numbers of the given accs
# Paramters: A database handle, one or more accs
# Output: String
sub _current_version_numbers {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my @accs = @_;

    assert(wantarray || @accs <= 1, "Cannot call in scalar context with multiple acc arguments\n");

    my $query = QUERY::Cache->get($dbh, $class, "get_version");
    unless ($query) {
	$query = create_query("type"=>"SVSelect", 
			      "dbh"=>$dbh,
			      "table"=>$class->get_version_table(),
			      "return_field"=>"_version",
			      "where"=>"_acc=?_acc?");
	QUERY::Cache->set($dbh, $class, "get_version", $query);
    }


    my @results;
    for (@accs) {
	$query->execute("_acc"=>$_);
	push(@results, $query->fetch_values());
    }

    return wantarray ? @results : $results[0];
}

    
########################
# Name: _update_current_version
# Purpose: Dynamic method to update the current version number in the version table
# Paramters: None
# Output: None
sub _update_current_version {
    my $self = assert_dynamic(shift);
    my $class = ref($self);
    my $dbh = $self->get_dbh();

    my $current_version = $class->_current_version_numbers($dbh, $self->get_acc());
    return if (defined $current_version && $current_version > $self->get_version());
    
    my $query;
    if (defined $current_version) {
	$query = QUERY::Cache->get($dbh, $class, "update_version");
	unless ($query) {
	    $query = create_query("type"=>"Update", 
				  "dbh"=>$dbh,
				  "table"=>$class->get_version_table(),
				  "attributes"=>["_version"],
				  "where"=>"_acc=?_acc?");
	    QUERY::Cache->set($dbh, $class, "update_version", $query);
	}
	
    }
    else {
	$query = QUERY::Cache->get($dbh, $class, "insert_version");
	unless ($query) {
	    $query = create_query("type"=>"Insert",
				  "dbh"=>$dbh,
				  "table"=>$class->get_version_table(),
				  "attributes"=>["_acc","_version"]);
	    QUERY::Cache->set($dbh, $class, "insert_version", $query);
	}
    }

    $query->execute("_version"=>$self->get_version, "_acc"=>$self->get_acc());

}

#####################
# Name: _reset_version
# Purpose: Dynamic method to reset the version number based on the contents of the table
# Parameters: None
# Output: None
sub _reset_version {
    my $self = assert_dynamic(shift);
    my $class = ref($self);
    my $dbh = $self->get_dbh();

    my $table;
    my $query = QUERY::Cache->get($dbh, $class, "max_version");
    unless ($query) {
	$table = $class->class_name();
	$query = create_query("type"=>"SVSelect",
			      "dbh"=>$dbh,
			      "table"=>$table,
			      "return_field"=>"version",
			      "function_fields" => {"$table.version"=>"max"},
			      "where"=>"acc=?acc?",
			      "command"=>"group by acc");
	QUERY::Cache->get($dbh, $class, "max_version", $query);
    }
    $query->execute("acc"=>$self->get_acc());
    my ($value) = $query->fetch_values();

    $query = QUERY::Cache->get($dbh, $class, "update_version");
    unless ($query) {
	$query = create_query("type"=>"Update", 
			      "dbh"=>$dbh,
			      "table"=>$class->get_version_table(),
			      "attributes"=>["_version"],
			      "where"=>"_acc=?_acc?");
	QUERY::Cache->set($dbh, $class, "update_version", $query);
    }
    
    $query->execute("_acc"=>$self->get_acc(), "_version"=>$value);
}


########################
# Name: insert
# Purpose: Augment DBClass::insert
# Parameters: No extra parameters
# Output: None
sub insert {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my @objs = ref($proto) ? ($proto) : @_;

    for (@objs) {
	assert(!$_->get_version(), "Cannot insert a $class object with a defined version number");
	$_->set_version(($class->_current_version_numbers($_->get_dbh(), $_->get_acc()) || 0)+1);
    }
    $class->CLASSES::DBClass::insert(@objs);
    for (@objs) {
	$_->_update_current_version();
    }
}


#######################
# Name: remove
# Purpose: Augment the DBCLass remove
# Parameters: None
# Output: None
sub remove {
    my $self = assert_dynamic(shift);
    $self->CLASSES::DBClass::remove();
    $self->_reset_version();
}

######################
# Name: update
# Purpose: Augment DBClass::create
# Parameters: None
# Output: None
sub update {
    my $self = assert_dynamic(shift);
    $self->CLASSES::DBClass::update();
    $self->_reset_version();
}

#######################
# Name: get
# Purpose: Augment DBClass::get
# Parameters: In addition to DBClass::get parameters:
#             "all_versions" => bool -- return all versions if true, only the latest version if false (default)
# Output: Object or set of objects
sub get {
    my $class = assert_static(shift);
    my %params = @_;

    unless ($params{"all_versions"}) {	
	my $version_table = $class->get_version_table();
	my $table = $class->class_name();
	$params{"join_tables"} = exists $params{"join_tables"} ? [$version_table, @{$params{"join_tables"}}] : [$version_table];
	$params{"where"} = QUERY::Where->new([ref($params{"where"}) eq "ARRAY" ? @{$params{"where"}} : $params{"where"}, "$version_table._acc=$table.acc and $version_table._version=$table.version"]);
    }
    return $class->CLASSES::DBClass::get(%params);
}

######################
# Name: get_pk 
# Purpose: Modification of DBClass::get_pk to avoind version restriction
# Parameters: A database handle and a primary key
# Output: The object
sub get_pk {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $primary_key = shift;
    assert($primary_key, "Must provide a primary key\n");
    
    my %extra_params = @_;
    $extra_params{"saved_query"} = "versioned_saved_query" unless $extra_params{"saved_query"};
    $extra_params{"all_versions"} = 1 unless exists $extra_params{"all_versions"};
    return $class->CLASSES::DBClass::get_pk($dbh, $primary_key, %extra_params);
}

#####################
# Name: get_by_acc
# Purpose: (static) Get the current version with a given acc
# Parameters: database handle, acc
# Output: Output of CLASSES::DBClass::get
sub get_by_acc {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $acc = $class->parse_lsid($dbh, shift);
    my %params = @_;

    assert(!$params{where}, "Will not take a where statement");
    assert(!$params{join_tables}, "Will not take additional join tables");

    $params{"saved_query"} = "get_by_acc";
    $params{"where"} = QUERY::Where->new(["acc=?gbacc?"]);
    $params{"values"} = {"gbacc"=>$acc};

    return $class->get("dbh"=>$dbh, %params);
}

######################
# Name: current_version
# Purpose: (dynamic) Deterimine of this is the current version of the object, as recorded in the database
# Parameters: None
# Output: boolean
sub current_version {
    my $self = assert_dynamic(shift);
    return undef unless $self->is_in_db() && $self->in_synch();
    assert(defined $self->get_acc() && defined $self->get_version(), ref($self)." object must have an acc and version");
    my ($current_version) = ref($self)->_current_version_numbers($self->get_dbh(), $self->get_acc());
    return defined $current_version && $self->get_version() == $current_version;
}

######################
# Name: correct_attribute
# Purpose: (dynamic) Correct the value of an attibute (without resetting the id or version)
# Parameters: attribute name, new value
# Output: None
sub correct_attribute {
    my $self = assert_dynamic(shift);
    my ($att, $value) = @_;
    my $id = $self->get_id();
    my $version = $self->get_version();
    my $fun = "set_$att";
    $self->$fun($value);
    $self->set_version($version);
    $self->set_id($id);
}

#######################
# Name: _reset_version
# Purpose: (dynamic) Change the version number
# Parameters: new version value
# Output: None
sub _modify_version {
    my $self = assert_dynamic(shift);
    my $class = ref($self);
    my $new_version = shift;

    $self->_modify_object_in_db("version" => $new_version);
    $self->_reset_version();
}

#######################
# Name: get_attribute_array
# Purpose: (dynamic) Print a string of all values
# Paramters: list of attributes to print (default = all)
# Output: A formated string
sub get_attribute_array {
   my $self = assert_dynamic(shift);
   my @atts = @_ > 0 ? @_ : sort {return -1 if $a eq "id" || $b eq "entry_date"; return 1 if $b eq "id" || $a eq "entry_date"; return -1 if $a eq "acc"; return 1 if $b eq "acc"; return -1 if $a eq "version"; return 1 if $b eq "version"; return -1 if $a le $b; return 1 if $b le $a; return 0} ref($self)->get_attribute_list();
   return $self->CLASSES::DBClass::get_attribute_array(@atts);
}

######################
# Name: mark_dead
# Purpose: (static) Mark acc "dead" (e.g. there is no "current" version)
# Parameters: Database handle, acc
# Output: None
sub mark_dead {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $acc = shift;

    my $version_table = $class->get_version_table();
    my $query = QUERY::Cache->get($dbh, $class, "mark_dead");
    unless ($query) {
	$query = create_query("type"=>"Remove", 
			      "dbh"=>$dbh, 
			      "table"=>$version_table, 
			      "where"=>"$version_table._acc=?acc?");
	QUERY::Cache->set($dbh, $class, "mark_dead", $query);
    }
    $query->execute("acc"=>$acc);
}

######################
# Name: reset_id
# Purpose: (dynamic) Reset the objects id to udef
# Parameters: None
# Output: None
sub reset_id {
    my $self = assert_dynamic(shift);
    $self->set_id(undef);
    $self->set_version(undef);
}

######################
# Name: class_copy_exact
# Purpose: (static) Class copy, preserving id and version
# Parameter: Operand object, operator object
# Output: None
sub class_copy_exact {
    my $class = assert_static(shift);
    my ($lvalue, $rvalue) = @_;

    $class->class_copy($lvalue, $rvalue);
    $lvalue->set_version($rvalue->get_version());
    $lvalue->set_id($rvalue->get_id());
}

###################
# Name: get_lsid
# Purpose (dynamic): Return acc (LSID format)
# Parameters: None
# Output: None
sub get_lsid {
    my $self = assert_dynamic(shift);
    my $class = ref($self);

    my $class_name = $class->class_name();
    my $acc = $self->get_acc();
    my $tax_id = CLASSES::Build->current_build($self->get_dbh())->get_tax_id();

    return "urn:lsid:pseudogene.org:$tax_id\.$class_name:$acc";
}

###################
# Name: set_lsid
# Purpose (dynamic): Set lsid_acc (parse out LSID format, if necessary)
# Parameters: acc (number or LSID format)
# Output: None
sub set_lsid {
    my $self = assert_dynamic(shift);
    my $class = ref($self);
    my $acc = assert(shift, "No accession number provided");

    $self->set_lsid_acc($self->parse_lsid($acc));
}

#################
# Name: parse_lsid_acc
# Purpose (static/dynamic): Validate an acc number and parse out the numberical portion
# Parameters: A dbh (if static), and acc
# Output: An integer
sub parse_lsid {
    my $self = shift;
    my $class;
    my $dbh;
    if (ref($self)) {
	$dbh = $self->get_dbh();
	$class = ref($self);
    }
    else {
	($class, $dbh) = ($self, assert_dbh(shift));
    }
    my $acc = assert(shift, "No accession number porvided");

    my $class_name = $class->class_name();
    my $tax_id = CLASSES::Build->current_build($dbh)->get_tax_id();

    if ($acc =~ /^urn:lsid:pseudogene.org:($tax_id\.){0,1}(CLASSES::){0,1}$class_name:([^:]+)$/) {
	return $3;
    }
    
    if ($acc =~ /^[^:]+$/) {
	return $acc;
    }

    assert(0, "Bad accession number format: $acc");
}


1;
