package CLASSES::Fragment;
use strict;
use CLASSES::DBClass_Versioned_Free;
use CLASSES::Chromosome;
use CLASSES::Chromosome_Segment;
use QUERY::query_factory;
use TOOLS::assert;

our @VERSION = '1.0';
our @ISA = qw(CLASSES::DBClass_Versioned_Free);
      

#-------------------------------------------------------------------------
# Required Functions
sub _vinit {
    my $self = shift;

    $self->set_strand("+") unless $self->get_strand();

    return $self;
}

#-------------------------------------------------------------------------
# Other methods

######################
# Name: set_start_seg
# Purpose: Overide the standard get_start_seg to reset _segment_list
# Parameters: New value
# Output: None
sub set_start_seg {
    my $self = assert_dynamic(shift);
    $self->{"_start_seg"} = $_[0];
    $self->{"_segment_list"} = undef;
    $self->{"_chromosome"} = undef;
    $self->{"_coords"} = undef;
    $self->{"_sequence"} = undef;
    $self->set_synch(undef);
    $self->set_version(undef);
    $self->set_id(undef);
}

######################
# Name: set_finish_seg
# Purpose: Overide the standard get_finish_seg to reset _segment_list
# Parameters: New value
# Output: None
sub set_finish_seg {
    my $self = assert_dynamic(shift);
    $self->{"_finish_seg"} = $_[0];
    $self->{"_segment_list"} = undef;
    $self->{"_chromosome"} = undef;
    $self->{"_coords"} = undef;
    $self->{"_sequence"} = undef;
    $self->set_synch(undef);
    $self->set_version(undef);
    $self->set_id(undef);
}

######################
# Name: get_segments
# Purpose: (dynamic) Get this array of chromosome segments used by the fragment
# Parameters: None
# Output: A list of segment objects
sub get_segments {
    my $self = assert_dynamic(shift);

    unless ($self->{"_segment_list"}) {
	my $first = CLASSES::Chromosome_Segment->get_pk($self->get_dbh(), $self->get_start_seg());
	if ($self->get_finish_seg() == $self->get_start_seg()) {
	    $self->{"_segment_list"} = [$first];
	}
	else {
	    my $last = CLASSES::Chromosome_Segment->get_pk($self->get_dbh(), $self->get_finish_seg());
	    my $chr = $first->get_chromosome();
	    my @arr = ($first);
	    for (my $i=$first->get_position()+1; $i < $last->get_position(); $i++) {
		push(@arr, $chr->get_segment_by_position($i));
	    }
	    $self->{"_segment_list"} = [@arr, $last];
	}
    }

    return @{$self->{"_segment_list"}};
}

#####################
# Name: get_first_segment
# Purpose: (dynamic) Get the first segment used by the fragment
# Parameters: None
# Output: The segment object
sub get_first_segment {
    my $self = assert_dynamic(shift);
    return ($self->get_segments())[0];
}

#####################
# Name: get_last_segment
# Purpose: (dynamic) Get the first segment used by the fragment
# Parameters: None
# Output: The segment object
sub get_last_segment {
    my $self = assert_dynamic(shift);
    return ($self->get_segments())[-1];
}

#####################
# Name: get_num_segments
# Purpose: (dynamic) Get the number of fragments used by the chromosome
# Parameters: None
# Output: An integer
sub get_num_segments {
    my $self = assert_dynamic(shift);
    return scalar $self->get_segments();
}

#####################
# Name: get_coords
# Purpose: (dynamic) Get the end-point coordinates of the fragment (relative to positive strand)
# Parameters: None
# Ouput: An array of the two coordinates
sub get_coords {
    my $self = assert_dynamic(shift);
    unless ($self->{"_coords"}) {
	my ($first, $last) = ($self->get_first_segment(), $self->get_last_segment());
	$self->{"_coords"} = [$first->offset_to_absolute($self->get_start_offset()), $last->offset_to_absolute($self->get_finish_offset())];
    }

    return @{$self->{"_coords"}};
}

######################
# Name: get_chomosome
# Purpose: (dynamic) Get the fragment's chromosome object
# Parameters: None
# Ouput: A chromosome object
sub get_chromosome {
    my $self = assert_dynamic(shift);
    unless ($self->{"_chromosome"}) {
	$self->{"_chromosome"} = $self->get_first_segment()->get_chromosome();
    }
    return $self->{"_chromosome"};
}


{ 
    #####################
    # Static variables for CLASSES::Fragment::get_by_location
	
    my @join_tables = ("Chromosome_Segment as StartSeg", "Chromosome_Segment as FinishSeg");
	
    my $chromosome_where = "StartSeg.chr_id=?chr_id? and FinishSeg.chr_id=?chr_id?";
    my $endpoint_where = "StartSeg.id=Fragment.start_seg and FinishSeg.id=Fragment.finish_seg";
    
    my $left_end_before_start = "(StartSeg.position < ?gbl_left_position?) or (StartSeg.position = ?gbl_left_position? and Fragment.start_offset < ?gbl_left?)";
    my $left_end_before_finish = "(StartSeg.position < ?gbl_right_position?) or (StartSeg.position = ?gbl_right_position? and Fragment.start_offset <= ?gbl_right?)";
    my $right_end_after_start = "(FinishSeg.position > ?gbl_left_position?) or (FinishSeg.position = ?gbl_left_position? and Fragment.finish_offset >= ?gbl_left?)";
    
    my $where_str = "(($left_end_before_start) and ($right_end_after_start)) or ((not ($left_end_before_start)) and ($left_end_before_finish))";
    my $gbl_where = QUERY::Where->new([$chromosome_where, $endpoint_where, $where_str]);
     
    ######################
    # Name: get_by_location
    # Purpose: (static) Get all fragments intersecting a given interval
    # Parameters: A database handle, a chromosome object, and coordinates on the 
    #             chromosome.  (No coords = whole chromosome, one coord = 
    #             single point interval)
    # Output: An array of fragments
    sub get_by_location {
	my $class = assert_static(shift);
	my $dbh = assert_dbh(shift);
	my $chr = shift;  assert(ref($chr) eq "CLASSES::Chromosome", "Not a chromosome: $chr\n");
	my ($left, $right) = @_==0 ? (1, $chr->get_size()) : @_==1 ? ($_[0], $_[0]) : ($_[0], $_[1]);
	my ($left_segment, $right_segment) = ($chr->get_segments($left), $chr->get_segments($right));
	my ($left_offset, $right_offset) = ($left_segment->absolute_to_offset($left), $right_segment->absolute_to_offset($right));
	
	my @arr = $class->get("dbh"=>$dbh,
			      "join_tables"=>\@join_tables,
			      "where"=>$gbl_where,
			      "values"=>{"gbl_left_position"=>$left_segment->get_position(),
					 "gbl_right_position"=>$right_segment->get_position(),
					 "gbl_left"=>$left_offset,
					 "gbl_right"=>$right_offset,
					 "chr_id" => $chr->get_id()},
			      "saved_query"=>"get_by_location");
	return @arr;
    }

}

######################
# Name: class_copy
# Purpose: Static method to copy the contents of the second object into the first object
# Parameter: Operand object, operator object
# Output: None
sub class_copy {
    my $class = assert_static(shift);
    my ($lvalue, $rvalue) = @_;
    $lvalue->{"_segment_list"} = $rvalue->{"_segment_list"};
    $class->CLASSES::DBClass_Versioned_Free::class_copy($lvalue, $rvalue);
}

######################
# Name: get_sequence
# Purpose: (dynamic) Print the fragment's sequence
# Parameters: None
# Output: A string
sub get_sequence {
    my $self = assert_dynamic(shift);

    unless ($self->{"_sequence"}) {
	my $C = $self->get_chromosome();
	$self->{"_sequence"} = $C->get_sequence($self->get_coords(), $self->get_strand());
    }
    return $self->{"_sequence"};
}

######################
# Name: get_length
# Purpose: (dynamic) Get the length of the fragment
# Parameters: None
# Output: The length
sub get_length {
    my $self = assert_dynamic(shift);

    return length($self->get_sequence());
}

######################
# Name: gtf_string
# Purpose: (dynamic) Print the fragment's gtf-format string
# Parameters: None
# Output: The string
sub gtf_string {
    my $self = assert_dynamic(shift);
    my @arr = map {$_ ? $_ : "."} ($self->get_id(), $self->get_acc(), $self->get_version(), $self->get_chromosome()->get_number(), $self->get_coords(), $self->get_strand());
    return "chr$arr[3]\t.\t.\t$arr[4]\t$arr[5]\t.\t$arr[6]\tacc $arr[1]; version $arr[2]; id $arr[0]";
}

######################
# Name: set_location
# Purpose: (dynamic) Set the coordinates of the fragment
# Parameters: The chromosome object and absolute coordinates (start->finish, 5' strand), strand direction (+/-)
# Output: 
sub set_location {
    my $self = assert_dynamic(shift);
    my ($C, $start, $finish, $strand) = @_;
    assert($C && ref($C) eq "CLASSES::Chromosome" && $start && $finish, "Bad parameters");
    assert($strand, "Must provide a strand (+/-)");

    my ($start_seg, $start_offset) = $C->get_segment_offset($start);
    my ($finish_seg, $finish_offset) = $C->get_segment_offset($finish);
    $self->set_start_seg($start_seg->get_id());
    $self->set_start_offset($start_offset);
    $self->set_finish_seg($finish_seg->get_id());
    $self->set_finish_offset($finish_offset);
    $self->set_strand($strand);
    $self->{"_coords"} = [$start, $finish];
}

1;
