package CLASSES::PFSet;
use strict;
use CLASSES::DBClass_Versioned_Free;
use CLASSES::Pseudogene;
use CLASSES::Fragment;
use TOOLS::assert;
use QUERY::Cache;
use QUERY::query_factory;

our @VERSION = '1.0';
our @ISA = qw(CLASSES::DBClass_Versioned_Free);

#-------------------------------------------------------------------------
# Required Functions
sub _vinit {
    my $self = shift;
    my %params = @_;

    $self->_set_content($params{"pgene_list"}, $params{"fragment_list"}, 1);

    return $self;
}

#-------------------------------------------------------------------------
# Class specific functions

######################
# Name: _set_content
# Purpose: Set the set content (pgenes and fragments)
# Parameters: A list (reference) of pgenes and a list (reference) of fragments
# Output: none
sub _set_content {
    my $self = assert_dynamic(shift);
    my $pgene_ref = shift;
    my $frag_ref =  shift;

    $self->{"_pgene_ref"} = $pgene_ref || [];
    $self->{"_frag_ref"} = $frag_ref || [];

    unless (shift @_) {
	$self->set_synch(undef);
	$self->set_id(undef);
	$self->set_version(undef);
    }
}

######################
# Name: get_fragments
# Purpose: (dynamic) Get the list of fragments
# Parameters: None
# Output: A list of fragment objects
sub get_fragments {
    my $self = assert_dynamic(shift);
    $self->_retrieve_content() unless defined $self->{_frag_ref};
    return @{$self->{"_frag_ref"}};
}

######################
# Name: get_pgenes
# Purpose: (dynamic) Get the list of pgenes
# Parameters: None
# Output: A list of pgene objects
sub get_pgenes {
    my $self = assert_dynamic(shift);
#    $self->_retrieve_content() unless defined $self->{_pgene_ref} or scalar(@{$self->{_pgene_ref}}) == 0;
    $self->_retrieve_content() if not defined $self->{_pgene_ref} or scalar(@{$self->{_pgene_ref}}) == 0;
    return @{$self->{"_pgene_ref"}};
}

######################
# Name: _retrieve_content
# Purpose: (dynamic)  Get the fragment and pgene content of the set from the db
# Parameters: None 
# Output: None
sub _retrieve_content {
    my $self = assert_dynamic(shift);
    my $class = ref($self);
    my $dbh = $self->get_dbh();

    my $query = QUERY::Cache->get($dbh, $class, "get-fragments");
    unless ($query) {
	$query = create_query(type         => "SVSelect",
			      dbh          => $dbh,
			      table        => "PFSet_to_Fragment",
			      return_field => "_Fragment_ID",
			      where        => "_PFSet_ID = ?id?");
	QUERY::Cache->set($dbh, $class, "get-fragments", $query);
    }
    $query->execute(id => $self->get_id());
    my @frag_arr = map {CLASSES::Fragment->get_pk($dbh, $_)} $query->fetch_values();

    $query = QUERY::Cache->get($dbh, $class, "get-pgenes");
    unless ($query) {
	$query = create_query(type         => "SVSelect",
			      dbh          => $dbh,
			      table        => "PFSet_to_Pgene",
			      return_field => "_Pgene_ID",
			      where        => "_PFSet_ID = ?id?");
	QUERY::Cache->set($dbh, $class, "get-pgenes", $query);
    }
    $query->execute(id => $self->get_id());
    my @pgene_arr = map {CLASSES::Pseudogene->get_pk($dbh, $_)} $query->fetch_values();
    $self->_set_content(\@pgene_arr, \@frag_arr, 1);
}

######################
# Name: _save_content
# Purpose: (dynamic) Saves the content to the database.  Set must be empty -- should only be called from insert
# Parameters: None
# Output: None
sub _save_content {
    my $self = assert_dynamic(shift);
    my $class = ref($self);
    my $dbh = $self->get_dbh();

    my $query = QUERY::Cache->get($dbh, $class, "save-fragments");
    unless ($query) {
	$query = create_query(type => "Insert",
			      dbh => $dbh,
			      table => "PFSet_to_Fragment",
			      attributes => ["_PFSet_ID","_Fragment_ID"]);
	QUERY::Cache->set($dbh, $class, "save-fragments", $query);
    }
    for ($self->get_fragments()) {
	$query->execute(_PFSet_ID => $self->get_id(), _Fragment_ID => $_->get_id());
    }

    $query = QUERY::Cache->get($dbh, $class, "save-pgenes");
    unless ($query) {
	$query = create_query(type => "Insert",
			      dbh => $dbh,
			      table => "PFSet_to_Pgene",
			      attributes => ["_PFSet_ID","_Pgene_ID"]);
	QUERY::Cache->set($dbh, $class, "save-pgenes", $query);
    }
    for ($self->get_pgenes()) {
	$query->execute(_PFSet_ID => $self->get_id(), _Pgene_ID => $_->get_id());
    }
}

######################
# Name: _delete_content
# Purpose: (dynamic) Saves the content to the database.  Set must be empty -- should only be called from insert
# Parameters: None
# Output: None
sub _delete_content {
    my $self = assert_dynamic(shift);
    my $class = ref($self);
    my $dbh = $self->get_dbh();

    my $query = QUERY::Cache->get($dbh, $class, "remove-fragments");
    unless ($query) {
	$query = create_query(type => "Remove",
			      dbh => $dbh,
			      table => "PFSet_to_Fragment",
			      where => "_PFSet_ID=?id?");
	QUERY::Cache->set($dbh, $class, "remove-fragments", $query);
    }
    for ($self->get_fragments()) {
	$query->execute(id => $self->get_id());
    }

    $query = QUERY::Cache->get($dbh, $class, "remove-pgenes");
    unless ($query) {
	$query = create_query(type => "Remove",
			      dbh => $dbh,
			      table => "PFSet_to_Pgene",
			      where => "_PFSet_ID=?id?");
	QUERY::Cache->set($dbh, $class, "remove-pgenes", $query);
    }
    for ($self->get_pgenes()) {
	$query->execute(id => $self->get_id());
    }
}

######################
# Name: insert
# Purpose: Augment DBClass::Versioned::Free to insert content
# Parameters: No extra parameters
# Output: None
sub insert {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my @objs = ref($proto) ? ($proto) : @_;

    for (@objs) {
	$_->set_size($_->size());
    }

    $class->CLASSES::DBClass_Versioned_Free::insert(@objs);
    for (@objs) {
	$_->_save_content();
    }
}

######################
# Name: get
# Purpose: Augement DBClass::Version::Free::get method to retrieve content
# Parameters: No extra parameters
# Output: None
sub get {
    my $class = assert_static(shift);
    my %params = @_;

    return $class->CLASSES::DBClass_versioned_Free::get(%params) if $params{get_components};

    my @objs = $class->CLASSES::DBClass_Versioned_Free::get(%params);
    my $dbh = @objs > 0 ? $objs[0]->get_dbh() : return wantarray ? () : undef;

    for (@objs) {
#	$_->_retrieve_content();
	$_->{_pgene_ref} = undef;     # Lazy evaluation
	$_->{_frag_ref} = undef;      # Lazy evaluation
    }

    return wantarray ? @objs: shift @objs;
}

######################
# Name: remove
# Purpose: Augment the DBClass_Versioned_Free::remove method  to remove content
# Parameters: No additional parameters
# Output: None
sub remove {
    my $self = assert_dynamic(shift);

    $self->_delete_content();
    $self->CLASSES::DBClass_Versioned_Free::remove();
}

######################
# Name: get_by_name
# Purpose: (static) Search the set by name
# Parameters: A database handle and a MySQL regular expression to search on
# Output: Any set containig the search word
sub get_by_name {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $str = shift;


    return $class->CLASSES::DBClass_Versioned_Free::get(dbh => $dbh,
						       where => "name like ?str?", 
						       "values"=>{str=>$str}, 
						       "saved_query"=>"get_by_name");
}

###############
# Name: get_by_id
# Purpose: (static) Search the set by id
# Parameters: A database handle and a set id to be searched on
# Output: Any set having this id 
sub get_by_id {
	my $class = assert_static(shift);
	my $dbh = assert_dbh(shift);
	my $set_id_search = shift;

	return $class->CLASSES::DBClass_Versioned_Free::get(
		dbh => $dbh,
		where => "id = ?id?",
		"values" => {id => $set_id_search},
		"saved_query" => "get_by_id"	
	);
}

###############
# Name: get_by_acc
# Purpose: (static) Search the set by acc
# Parameters: A database handle and a set acc to be searched on
# Output: Any imminent set having this acc 
sub get_by_acc {
	my $class = assert_static(shift);
	my $dbh = assert_dbh(shift);
	my $set_acc_search = shift;

	return $class->CLASSES::DBClass_Versioned_Free::get(
		dbh => $dbh,
		where => "acc = ?acc?",
		"values" => {acc => $set_acc_search},
		"saved_query" => "get_by_acc"	
	);
}

######################
# Name: get_by_notes
# Purpose: (static) Search the set by name
# Parameters: A database handle and a MySQL regular expression to search on
# Output: Any set containig the search word
sub get_by_notes {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $str = shift;


    return $class->CLASSES::DBClass_Versioned_Free::get(dbh => $dbh,
						       where => "notes like ?str?", 
						       "values"=>{str=>$str}, 
						       "saved_query"=>"get_by_name");
}

######################
# Name: get_by_pgene_id
# Purpose: (static) Get all sets using all the specified pgenes
# Parameters: database handle, a pgene
# Output: A list of sets
sub get_by_pgene {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $pgene = shift;
    assert(ref($pgene) eq "CLASSES::Pseudogene", "Not a pseudogene");

    my $query = QUERY::Cache->get($dbh, $class, "get-by-pgene");
    unless ($query) {
	$query = create_query(type => "SVSelect",
			      dbh => $dbh,
			      table => "PFSet_to_Pgene",
			      return_field => "_PFSet_ID",
			      where => "_Pgene_ID = ?id?");
	QUERY::Cache->set($dbh, $class, "remove_fragments", $query);
    }
    $query->execute(id => $pgene->get_id());
    
    return map {CLASSES::PFSet->get_pk($dbh, $_)} $query->fetch_values();
}

######################
# Name: get_by_fragment
# Purpose: (static) Get all sets using the specified pgenes
# Parameters: database handle, a pgene
# Output: A list of sets
sub get_by_fragment {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $pgene = shift;
    assert(ref($pgene) eq "CLASSES::Fragment", "Not a fragment");

    my $query = QUERY::Cache->get($dbh, $class, "get-by-pgene");
    unless ($query) {
	$query = create_query(type => "SVSelect",
			      dbh => $dbh,
			      table => "PFSet_to_Fragment",
			      return_field => "_PFSet_ID",
			      command => "order by _PFSet_ID",
			      where => "_Fragment_ID = ?id?");
	QUERY::Cache->set($dbh, $class, "remove_fragments", $query);
    }
    $query->execute(id => $pgene->get_id());
    
    return map {CLASSES::PFSet->get_pk($dbh, $_)} $query->fetch_values();
}

######################
# Name: size
# Purpose: (dynamic) Number of objects in the set
# Parameters: Takes one paramter:
#             'frag': return number of fragments
#             'pgene': return number of pgenes
#             else: return total number of objects
# Output: An int
sub size {
    my $self = assert_dynamic(shift);
    my $param = shift;

    if ($param) {
	$param =~ s/s$//;
	if ($param eq 'frag' || $param eq 'fragment') {
	    return scalar $self->get_fragments();
	}
	elsif ($param eq 'pgene' || $param eq 'pseudogene') {
	    return scalar $self->get_pgenes();
	}
    }
    return (scalar $self->get_fragments()) + (scalar $self->get_pgenes());
}

1;

