package CLASSES::Pseudogene;
use strict;
use CLASSES::DBClass_Versioned_Free;
use CLASSES::Fragment;
use CLASSES::Sequence_Alignment;
use CLASSES::Protein;
use TOOLS::assert;
use TOOLS::translate;
use QUERY::query_factory;
use TaxID::tax_id;

our @VERSION = '1.0';
our @ISA = qw(CLASSES::DBClass_Versioned_Free);

#-------------------------------------------------------------------------
# Required Functions
sub _vinit {
    my $self = shift;
    my %params = @_;

    $self->{_extension_hash} = {};
    $self->correct_attribute("ptype", ".") unless $self->get_ptype();
    $self->correct_attribute("source", "pseudogene.org") unless $self->get_source();
    $self->set_fragment_info(@{$params{"fragment_info"}});
    $self->add_extension_attributes(@{$params{"extensions_attributes"}}, 1);

    return $self;
}

#-------------------------------------------------------------------------
# Extra functions

my $fragment_relation_table = TOOLS::constants::FRAGMENT_RELATION_TABLE;
my $ext_table = "Pseudogene_Extensions";

######################
# Name: get
# Purpose: Augment the DBClass_Versioned_Free::get method to extract fragment and extension information
# Parameters: No additional parameters
# Output: The array of objects
sub get {
    my $class = assert_static(shift);
    my %params = @_;

    return $class->CLASSES::DBClass_Versioned_Free::get(%params) if $params{get_components};

    my @objs = $class->CLASSES::DBClass_Versioned_Free::get(@_);
    my $dbh = @objs > 0 ? $objs[0]->get_dbh() : return wantarray ? () : undef;
    # @objs holds the list of pseudogenes -- now get the list of fragments and sequence alignments for each
    my $query = QUERY::Cache->get($dbh, $class, "get-fragments");
    unless ($query) {
	$query = create_query("type"=>"Select", 
			      "dbh"=>$dbh,
			      #"table_hash"=>{$fragment_relation_table=>["_Fragment_ID,_alignment"]},    # CHAGNED THIS (CORRECT?)
			      "table_hash"=>{$fragment_relation_table=>["_Fragment_ID","_alignment"]}, 
			      "where"=>["$fragment_relation_table._Pgene_ID=?pgene_id?"], 
			      "command"=>"order by $fragment_relation_table._sequence_position");
	QUERY::Cache->set($dbh, $class, "get-fragments", $query);
    }

    for (@objs) {
	next if $_->_get_fragment_info();
	assert($_->get_dbh==$dbh, "Varying database handles\n");
	$query->execute("pgene_id"=>$_->get_id());

	my @results;
	while (my $ref = $query->fetch_row()) {
	    push(@results, [CLASSES::Fragment->get_pk($dbh, $ref->{"_Fragment_ID"}), $ref->{"_alignment"} ? CLASSES::Sequence_Alignment->get_pk($dbh, $ref->{"_alignment"}) : undef]);
	}
	$query->finish();

	$_->set_fragment_info(@results);
	$_->set_synch('1');
    }

    for (@objs) {
	$_->_retrieve_extensions();
    }

    assert(wantarray || @objs<=1, "Can't call in scalar context when number of results is more than 1\n");

    return wantarray ? @objs : shift @objs;
}

######################
# Name: _add_fragments
# Purpose: (static) Add fragments to the table for each passed object
# Parameters: One or more pgenes
# Output: None
sub _add_fragments {
    my $class = assert_static(shift);
    my @objs = @_;
    my $dbh = $objs[0]->get_dbh();

    my $query = QUERY::Cache->get($dbh, $class, "insert-fragments");
    unless ($query) {
	$query = create_query("type"=>"Insert", 
			      "dbh"=>$dbh,
			      "table"=>$fragment_relation_table, 
			      "attributes"=>[qw(_Fragment_ID _Pgene_ID _sequence_position _last_frag _alignment)]);
	QUERY::Cache->set($dbh, $class, "insert-fragments", $query);
    }

    foreach my $pgene (@objs) {
	assert(ref($pgene) eq $class, "Bad object passed: $pgene\n");
	assert($pgene->get_dbh()==$dbh, "Varying database handles\n");
	my @arr = $pgene->_get_fragment_info();
	for (my $i=0; $i < @arr; $i++) {
	    $query->execute("_Pgene_ID" => $pgene->get_id(), 
			    "_Fragment_ID" => $arr[$i]->[0] ? $arr[$i]->[0]->get_id() : undef,
			    "_sequence_position" => $i+1,
			    "_alignment" => $arr[$i]->[1] ? $arr[$i]->[1]->get_id() : undef,
			    "_last_frag" => ($i==@arr-1) ? "T" : "F");
	}
    }
}

######################
# Name: _remove_fragments
# Purpose: (static) Remove each pgene and its fragments from the table
# Parameters: One or more pgenes
# Output: None
sub _remove_fragments {
    my $class = assert_static(shift);
    my @objs = @_;
    my $dbh = $objs[0]->get_dbh();

    my $query = QUERY::Cache->get($dbh, $class, "remove-fragments");
    unless ($query) {
	$query = create_query("type"=>"Remove", 
			      "dbh"=>$dbh,
			      "table"=>$fragment_relation_table, 
			      "where"=>"$fragment_relation_table._Pgene_ID=?pgene_id?");
	QUERY::Cache->set($dbh, $class, "remove-fragments", $query);
    }

    foreach my $pgene (@objs) {
	assert(ref($pgene) eq $class, "Bad object passed: $pgene\n");
	assert($pgene->get_dbh()==$dbh, "Varying database handles\n");
	$query->execute("pgene_id"=>$pgene->get_id());
    }
}

######################
# Name: insert
# Purpose: Augment the DBClass_Versioned_Free::insert method to insert fragment information
# Parameters: No additional parameters
# Output: None
# Note: Assumes fragments are sorted (as done in set_fragment_info)
sub insert {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my @objs = ref($proto) ? ($proto) : @_;

    foreach (@objs) {
	$_->get_GC_Pgene();
    }

    $class->CLASSES::DBClass_Versioned_Free::insert(@objs);
    $class->_add_fragments(@objs);
    for (@objs) {
	$_->_save_extensions();
    }
}

######################
# Name: update
# Purpose: Augment the DBClass_Versioned_Free::update method to update fragment information
# Parameters: No additional parameters
# Output: None
sub update {
    my $self = assert_dynamic(shift);
    return if $self->in_synch();

    $self->get_GC_Pgene();

    $self->CLASSES::DBClass_Versioned_Free::update();
    ref($self)->_remove_fragments($self);
    ref($self)->_add_fragments($self);
}

######################
# Name: remove
# Purpose: Augment the DBClass_Versioned_Free::remove method to remove fragment information
# Parameters: No additional parameters
# Output: None
sub remove {
    my $self = assert_dynamic(shift);
    ref($self)->_remove_fragments($self);
    $self->CLASSES::DBClass_Versioned_Free::remove();
}

######################
# Name: set_fragment_info
# Purpose: (dynamic) Set the fragment information (sorting them by position)
# Parameters: One or more fragment/alignment tuples
# Output: None
# Note: Assumes that all fragments are on the same chromosome (dosn't check for it)
sub set_fragment_info {
    my $self = assert_dynamic(shift);
    $self->{"_fragment_info"} = @_ > 1 ? [sort {($a->[0]->get_coords())[0] <=> ($b->[0]->get_coords())[0]} @_] : [@_];
    $self->{"_protein"} = undef;
    $self->{"_sequence"} = undef;
    $self->set_synch(undef);
#    $self->set_id(undef);
#    $self->set_version(undef);
}

######################
# Name: _get_fragment_info
# Purpose: (dynamic) Return the list of fragment/alignment information
# Parameters: None
# Output: A list of tuple references
sub _get_fragment_info {
    my $self = assert_dynamic(shift);
    return @{$self->{"_fragment_info"}};
}

######################
# Name: get_protein
# Purpose: (dynamic) Return the associated protein
# Parameters: None
# Output: The protein object
sub get_protein {
    my $self = assert_dynamic(shift);
    unless ($self->{"_protein"}) {
	my $alignment = $self->get_alignments(0) || return undef;
	$self->{"_protein"} = CLASSES::Protein->get_pk($self->get_dbh(), $alignment->get_protein_id());
    }

    return $self->{"_protein"};
}

######################
# Name: num_fragments
# Purpose: (dynamic) Return the number of fragments (exons)
# Parameters: None
# Output: An integer
sub num_fragments {
    my $self = assert_dynamic(shift);
    return scalar $self->_get_fragment_info();
}

######################
# Name: num_exons
# Purpose: (dynamic) Return the number of exons
# Parameters: None
# Output: An integer
sub num_exons {
    my $self = assert_dynamic(shift);
    return $self->num_fragments(@_);
}

######################
# Name: get_fragments
# Purpose: (dynamic) Return the ith (or all) pseduogene fragments
# Parameters: i (or undef) -- -1 for last fragment
# Output: A fragment (or list of fragments)
sub get_fragments {
    my $self = assert_dynamic(shift);
    my $i = shift;
    if (defined $i) {
	assert(-1 <= $i && $i < $self->num_fragments(), "Out of range: $i(".$self->num_fragments().")\n");
	return ($self->_get_fragment_info())[$i]->[0];
    }
    else {
	assert(wantarray, "Bad use in scalar context\n");
	return map {$_->[0]} $self->_get_fragment_info();
    }
}

######################
# Name: get_alginments
# Purpose: (dynamic) Return the ith (or all) alignments
# Parameters: i (or undef)
# Output: An alignment (or list of alignments)
sub get_alignments {
    my $self = assert_dynamic(shift);
    my $i = shift;
    if (defined $i) {
	assert(-1 <= $i && $i <= $self->num_fragments(), "Out of range: $i\n");
	return ($self->_get_fragment_info())[$i]->[1];
    }
    else {
	assert(wantarray, "Bad use in scalar context\n");
	return map {$_->[1]} $self->_get_fragment_info();
    }
}

######################
# Name: get_coords
# Purpose: (dynamic) Get the end-point coordinates of the fragment (relative to positive strand)
# Parameters: None
# Output: An array of the two coordinates
sub get_coords {
   my $self = assert_dynamic(shift);
   my ($first,$last) = ($self->get_fragments(0), $self->get_fragments(-1));
   return (($first->get_coords())[0], ($last->get_coords())[1]);
}

######################
# Name: get_length
# Purpose: (dynamic) Get the total length of the pgene (introns included)
# Parameters: None
# Output: An integer
sub get_length {
    my $self = assert_dynamic(shift);
    my @coords = $self->get_coords();
    return $coords[1]-$coords[0]+1;
}

######################
# Name: get_fragment_coords
# Purpose: (dynamic) Get the end-point coordinates of the exons
# Parameters: None
# Output: A list of array references, where the ith array contains the coordinates for the ith exon
sub get_fragment_coords {
    my $self = assert_dynamic(shift);
    my @frags = $self->get_fragments();
    return map {[$_->get_coords()]} @frags;
}

######################
# Name: get_coding_length
# Purpose: (dynamic) Get the total length of the "coding" exons
# Parameters: None
# Output: An integer
sub get_coding_length {
    my $self = assert_dynamic(shift);
    my @frags = $self->get_fragments();
    my $total = 0;
    for (@frags) {
	$total += $_->get_length();
    }
    return $total;
}


#######################
# Name: get_by_protein
# Purpose: (static) Get all pseudogenes associated with a given protein
# Parameters: A database handle and a protein ACC
# Output: An array of pseudogene objects
sub get_by_protein {
   my $class = assert_static(shift);
   my $dbh = assert_dbh(shift);
   my $acc = CLASSES::Protein->get_primary_acc($dbh, shift);
   my %params = @_;

   my @join_tables = ($fragment_relation_table, "Sequence_Alignment", "Protein");
   my @where = ("Pseudogene.id=$fragment_relation_table._pgene_id", "$fragment_relation_table._alignment=Sequence_Alignment.id", "Sequence_Alignment.protein_id=Protein.id", "Protein.acc=?pacc?");

   $params{saved_query} = "get_by_protein" unless ($params{saved_query} || $params{join_tables} || $params{where});
   $params{where} = $params{where} ? QUERY::Where->new([\@where, $params{where}]) : \@where;
   $params{join_tables} = $params{join_table} ? [@{$params{join_tables}}, @join_tables] : \@join_tables;

   return $class->get("dbh"=>$dbh,
		      "command"=>"group by Pseudogene.id",
		      "values"=>{"pacc"=>$acc},
		      %params);
}

#######################
# Name: get_by_gene
# Purpose: (static) Get all pseudogenes associated with a given gene
# Parameters: A database handle and an (ensembl) gene id
# Output: An array of pseudogene objects
sub get_by_gene {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $gene_id = assert(shift, "No gene id provided");
    my %params = @_;

    my @join_tables = ($fragment_relation_table, "Sequence_Alignment", "Protein");
    my @where = ("Pseudogene.id=$fragment_relation_table._pgene_id", "$fragment_relation_table._alignment=Sequence_Alignment.id", "Sequence_Alignment.protein_id=Protein.id", "Protein.gene_id=?gene_id?");

    $params{saved_query} = "get_by_gene" unless $params{saved_query} || $params{join_tables} || $params{where};
    $params{where} = $params{where} ? QUERY::Where->new([\@where, $params{where}]) : \@where;
    $params{join_tables} = $params{join_table} ? [@{$params{join_tables}}, @join_tables] : \@join_tables;
    
    return $class->get(dbh => $dbh,
		       values => {gene_id => $gene_id},
		       %params);
}

########################
# Name: get_by_source
# Purpose: (Static) Get all pseudogenes associated with a given source
# Parameters: A database handler and a source
# Output: An array of pseudogene objects 
sub get_by_source {
	my $class = assert_static(shift);
	my $dbh = assert_dbh(shift);
	my $source = shift;

	my @where = ("source=?src?");
	return $class->get(
		dbh => $dbh,
		where => \@where,
		values => {src => $source},
		saved_query => "get_by_source",
		all_versions => 1,
	);
} 


{ 
    #####################
    # Static variables for CLASSES::Pseudogene::get_by_location
	
    my @gbl_join_tables = ("Fragment_to_Pgene as Start", "Fragment as StartFragment", "Chromosome_Segment as StartSeg", "Fragment_to_Pgene as Finish", "Fragment as FinishFragment", "Chromosome_Segment as FinishSeg");
	
    my $chromosome_where = "StartSeg.chr_id=?chr_id? and FinishSeg.chr_id=?chr_id?";
    my $endpoint_where = "StartSeg.id=StartFragment.start_seg and FinishSeg.id=FinishFragment.finish_seg";
    my $start_fragment_where = "Pseudogene.id=Start._Pgene_ID and Start._sequence_position=1 and Start._Fragment_ID=StartFragment.id";
    my $finish_fragment_where = "Pseudogene.id=Finish._Pgene_ID and Finish._last_frag='T' and Finish._Fragment_ID=FinishFragment.id";
    my $left_end_before_start = "(StartSeg.position < ?gbl_left_position?) or (StartSeg.position = ?gbl_left_position? and StartFragment.start_offset < ?gbl_left?)";
    my $left_end_before_finish = "(StartSeg.position < ?gbl_right_position?) or (StartSeg.position = ?gbl_right_position? and StartFragment.start_offset <= ?gbl_right?)";
    my $right_end_after_start = "(FinishSeg.position > ?gbl_left_position?) or (FinishSeg.position = ?gbl_left_position? and FinishFragment.finish_offset >= ?gbl_left?)";
    
    my $where_str = "(($left_end_before_start) and ($right_end_after_start)) or ((not ($left_end_before_start)) and ($left_end_before_finish))";
    my $gbl_where = QUERY::Where->new([$chromosome_where, $endpoint_where, $start_fragment_where, $finish_fragment_where, $where_str]);
     
    ######################
    # Name: get_by_location
    # Purpose: (static) Get all fragments intersecting a given interval
    # Parameters: A database handle, a chromosome object, and coordinates on the 
    #             chromosome.  (Undef for both coords = whole chromosome, one coord = 
    #             single point interval). Other get parameters.
    # Output: An array of fragments
    sub get_by_location {
	my $class = assert_static(shift);
	my $dbh = assert_dbh(shift);
	my $chr = shift;  assert(ref($chr) eq "CLASSES::Chromosome", "Not a chromosome: $chr\n");

	my $left = shift;  $left = 1 unless $left && $left > -1;
	my $right = shift; $right = $chr->get_size() unless $right && $right > -1;

	my %params = @_;

	my ($left_segment, $right_segment) = ($chr->get_segments($left), $chr->get_segments($right));
	my ($left_offset, $right_offset) = ($left_segment->absolute_to_offset($left), $right_segment->absolute_to_offset($right));
	
	$params{saved_query} = $params{where} || $params{join_tables} ? undef : $params{save_query} || "get_by_location";
	$params{where} = $params{where} ? QUERY::Where->new([$gbl_where, $params{where}]) : $gbl_where;
	$params{join_tables} = $params{join_tables} ? [@{$params{join_tables}}, @gbl_join_tables] : \@gbl_join_tables;

	my @arr = $class->get("dbh"=>$dbh,
			      #"where"=> $where,
			      "values"=>{"gbl_left_position"=>$left_segment->get_position(),
					 "gbl_right_position"=>$right_segment->get_position(),
					 "gbl_left"=>$left_offset,
					 "gbl_right"=>$right_offset,
					 "chr_id" => $chr->get_id()},
			      %params);   # %params covers join_tables, where and saved_query
	return @arr;
    }
}


######################
# Name: get_by_fragment
# Purpose: Get all pseudogenes using a given fragment
# Parameters: A database handle, a fragment object
# Output: An array of pseudogenes
sub get_by_fragment {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $fragment = shift;
    my %params = @_;
    assert(ref($fragment) eq "CLASSES::Fragment", "Bad fragment: $fragment");

    my @join_tables = ("Fragment_to_Pgene");
    my $where_str = "Fragment_to_Pgene._Fragment_ID=?gbf_frag_id? and Fragment_to_Pgene._Pgene_ID=Pseudogene.id";

    $params{saved_query} = "get_by_fragment" unless $params{saved_query} || $params{join_tables} || $params{where};
    $params{where} = $params{where} ? QUERY::Where->new([$where_str, $params{where}]) : $where_str;
    $params{join_tables} = $params{join_table} ? [@{$params{join_tables}}, @join_tables] : \@join_tables;

    my @arr = CLASSES::Pseudogene->get(dbh => $dbh, 
				       values => {"gbf_frag_id"=>$fragment->get_id()}, 
				       %params);
    return @arr;
}

######################
# Name: get_by_fragment_acc
# Purpose: Get all pseudogenes using a given framgnet acc (regardless of version)
# Parameters: A database handle, a fragment object
# Output: An array of pseudogenes
sub get_by_fragment_acc {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $fragment_acc = shift;
    my %params = @_;
    assert(!ref($fragment_acc), "Bad fragment: $fragment_acc");

    my @join_tables = ("Fragment_to_Pgene", "Fragment");
    my $where_str = "Fragment.acc=?gbfa_acc? and Fragment_to_Pgene._Fragment_ID=Fragment.id and Fragment_to_Pgene._Pgene_ID=Pseudogene.id";

    $params{saved_query} = "get_by_fragment_acc" unless $params{saved_query} || $params{join_tables} || $params{where};
    $params{where} = $params{where} ? QUERY::Where->new([$where_str, $params{where}]) : $where_str;
    $params{join_tables} = $params{join_table} ? [@{$params{join_tables}}, @join_tables] : \@join_tables;

    my @arr = CLASSES::Pseudogene->get("dbh"=>$dbh, "values"=>{"gbfa_acc"=>$fragment_acc}, %params);
    return @arr;
}

######################
# Name: class_copy
# Purpose: Static method to copy the contents of the second object into the first object
# Parameter: Operand object, operator object
# Output: None
sub class_copy {
    my $class = assert_static(shift);
    my ($lvalue, $rvalue) = @_;
    $lvalue->{"_fragment_info"} = $rvalue->{"_fragment_info"};
    $lvalue->{"_protein"} = $rvalue->{"_protein"};
    $lvalue->{_extension_hash} = $rvalue->{_extension_hash};
	
    $class->CLASSES::DBClass_Versioned_Free::class_copy($lvalue, $rvalue);    
}

######################
# Name: get_chromosome
# Purpose: (Dynamic) Get the chromosome object
# Parameters: None
# Output: A chromosomoe object
sub get_chromosome {
    my $self = assert_dynamic(shift);
    return $self->get_fragments(0)->get_chromosome();
}

######################
# Name: get_strand
# Purpose: (Dynamic) Get the strand 
# Parameters: None
# Output: +/-
sub get_strand {
    my $self = assert_dynamic(shift);
    return $self->get_fragments(0)->get_strand();
}

######################
# Name: get_sequence
# Purpose: (dynamic) Get the sequence
# Parameters: i (optional), numbered from 0
# Output: A string representing the ith exon, or the whole pseudogene if i=undef
sub get_sequence {
    my $self = assert_dynamic(shift);
    my $i = shift;

    if (defined $i) {
	$self->get_fragments(0)->get_sequence();
    }
    else {
	unless ($self->{"_get_sequence"}) {
	    my $C = $self->get_chromosome();
	    $self->{"_get_sequence"} = $C->get_sequence($self->get_coords(), $self->get_strand());
	}
	return $self->{"_get_sequence"};
    }
}

######################
# Name: gtf_string
# Purpose: (dynamic) Print the gtf string representing the pgene
# Parameters: None
# Output: The gtf_string
sub gtf_string {
    
    my $self = assert_dynamic(shift);
    my ($chr, $strand, $type, $protein, $acc, $version, $id) = map {$_ ? $_ : "."} ($self->get_chromosome()->get_number(), $self->get_strand(), $self->get_ptype(), $self->get_protein(), $self->get_acc(), $self->get_version(), $self->get_id()); # chr, start, finish, strand, type, protein, acc, version, id
    my @frags = $self->get_fragments();
#    my $exon_str = @frags.":".join(":", map {join("-", $_->get_coords())} @frags);
    my $protein_acc = ref($protein) && ref($protein) eq 'CLASSES::Protein'  ? $protein->get_acc() : ".";

    my $source = $self->get_source();

    my $name = $self->pgene_name();

    my $tax_id = $self->tax_id();
    my $trans_id = $self->get_lsid();
    my $gene_id = $self->get_id();

    my $str = "";

    my $tab = "\t";
    for (my $i=0; $i < @frags; $i++) {
	my ($start, $finish) = $frags[$i]->get_coords();
	$str = $str."chr$chr".$tab.$source.$tab."pseudogene ($type)".$tab.$start.$tab.$finish.$tab.".".$tab.$strand.$tab.".".$tab."pgene_type $type; protein $protein_acc; exon_index ".($i+1)."; name $name; tax_id $tax_id; gene_id $gene_id; transcript_id $trans_id\n";
    }
    return $str;
}   

######################
# Name: pgene_name
# Purpose: (dynamic) Return the pgene.org name
# Parameters: None
# Output: String
sub pgene_name {
    my $self = assert_dynamic(shift);
    my $dbh = $self->get_dbh();

    my $tax_id = $self->tax_id();
    my $org = get_org($tax_id);
    my $protein = $self->get_protein();
    my $protein_name = $protein ? ($protein->get_name() || $protein->get_acc()) : "unknown";
    my $chr_num = $self->get_chromosome()->get_number();
    $protein_name =~ s/ /\_/g;
    my $start = ($self->get_coords())[0];
    my $megabase = int($start/1000000);

    my $v = undef;
    if ($protein) {
	my @overlaps = CLASSES::Pseudogene->get_by_location($dbh, $self->get_chromosome(), $megabase, $start, join_tables => [$fragment_relation_table, "Sequence_Alignment", "Protein"], where=>QUERY::Where->new(clauses=>["Pseudogene.id=$fragment_relation_table._pgene_id", "$fragment_relation_table._alignment=Sequence_Alignment.id", "Sequence_Alignment.protein_id=Protein.id", "Protein.acc=?pacc?"]), values=>{pacc=>$protein->get_acc()});
	$v = scalar(@overlaps);
    }

    return join("_", split(/\s/, "$protein_name.$org.chr$chr_num.mb$megabase")).($v ? ".$v" : "");
}

######################
# Name: pgene_id
# Purpose: (dynamic) Return the pgene.org id
# Parameters: None
# Output: String
sub pgene_id {
    my $self = assert_dynamic(shift);

    my $id = $self->get_id();
    my $tax_id = $self->tax_id();
    return "PGO.$tax_id.$id";
}

#######################
# Name: get_fragment_sequence
# Purpose: (dynamic) Get the putative transcribed dna sequence of the pgene
# Parameters: None
# Output: The dna sequence.
sub get_fragment_sequence{
	my $self = assert_dynamic(shift);
	my $results = "";
	my @alignments = $self->get_alignments();

	my $found = 1;
    for (@alignments) {
	if (!$_) {
	    $found = 0;
	    last;
	}
	my $frag = $_->get_fragment();
	my $seq = $frag->get_sequence();
	if (!$seq) {
	    return "unassigned";
	}
	$results .= $seq;
    }

    return $found ? $results : "unknown";

}
######################
# Name: get_aligned_translation
# Purpose: (dynamic) Get the putative transcribed product of the pgene (account for frame shifts)
# Parameters: None
# Output: The spliced translation
sub get_aligned_translation {
    my $self = assert_dynamic(shift);

    my $results = "";
    my @alignments = $self->get_alignments();

    my $found = 1;
    for (@alignments) {
	if (!$_) {
	    $found = 0;
	    last;
	}
	my $seq = $_->get_frag_align_seq();
	if (!$seq) {
	    return "unassigned";
	}
	$results .= $seq;
    }

    return $found ? $results : "unknown";
}

######################
# Name: get_translation
# Purpose: (dynamic) Get the putative transcrived product of the pegene
# Parameters: None 
# Output: The spliced translation
sub get_translation {
    my $self = assert_dynamic(shift);

    my $results = "";
    for (my $i=0; $i  < $self->num_exons(); $i++) {
	$results .= translate($self->get_sequence($i));
    }

    return $results;
}

#######################
# Name: get_current_version
# Purpose: (dynamic) Get the current verion pseudogene belong to the input
# Parameters: database handle, id
# Output: id (current version)

sub get_current_version{
	my $self = assert_dynamic(shift);
	my $class = ref($self);
	my $dbh = $self->get_dbh();
	my $acc = $self->get_acc();

    	return $class->CLASSES::DBClass_Versioned_Free::get_by_acc($dbh, $acc, @_);
}

######################
# Name: get_by_acc
# Purpose: (static) Modifies DBClass_Versioned::get_by_acc to account for the extra pgene acc info.
# Parameters: database handle, acc
# Output: Output of CLASSES::DBClass::get
sub get_by_acc {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $acc = shift;

    $acc =~ s/^PGO\.\d+\.//;

    return $class->CLASSES::DBClass_Versioned_Free::get_by_acc($dbh, $acc, @_);
}


######################
# Name: (dynamic) _retrieve_extensions
# Purpose: Retrieve the extension hash table from the database
# Parameters: None
# Output: None
sub _retrieve_extensions {
    my $self = assert_dynamic(shift);
    my $class = ref($self);
    my $dbh = $self->get_dbh();

    my $query = QUERY::Cache->get($dbh, $class, "retrieve_extensions");
    unless (defined $query) {
	$query = create_query(type       => "Select",
			      dbh        => $dbh,
			      table_hash => {$ext_table => ["_att_name","_value"]},
			      where       => "_pgene_id = ?id?");
	QUERY::Cache->set($dbh, $class, "retrieve_extensions", $query);
    }
    $query->execute(id => $self->get_id());
    
    $self->{_extension_hash} = {};
    while (my $H = $query->fetch_row()) {
	$self->{_extension_hash}->{$H->{_att_name}} = $H->{_value};
    }
    $query->finish();
}

######################
# Name:(dynamic) _save_extensions
# Purpose: Save extensions to database
# Parameters: None
# Output: None
sub _save_extensions {
    my $self = assert_dynamic(shift);
    my $class = ref($self);
    my $dbh = $self->get_dbh();

    my $insert_query = QUERY::Cache->get($dbh, $class, "_save_insert");
    unless ($insert_query) {
	$insert_query = create_query(type  => "Insert",
				     dbh   => $dbh,
				     table => $ext_table,
				     attributes => ["_pgene_id", "_att_name", "_value"]);
        QUERY::Cache->set($dbh, $class, "_save_insert", $insert_query);
    }

    while (my ($att, $val) = each %{$self->{_extension_hash}}) {
	$insert_query->execute(_pgene_id=>$self->get_id(), _att_name=>$att, _value=>$val);
    }
}

######################
# Name: add_extension_attributes
# Purpose: (dynamic) Add extension attributes to the pgene
# Parameters: Array of (Name, value, name, value, ...); fianl parameter to prevent the id from getting reset
# Output: None
sub add_extension_attributes {
    my $self = assert_dynamic(shift);

    while (@_ > 1) {
	my ($att, $val) = (shift @_, shift @_);
	$self->{_extension_hash}->{$att} = $val;
    }

    if (@_ == 0) {
	$self->set_synch(undef);
	$self->set_id(undef);
	$self->set_version(undef);
    }
}

######################
# Name: get_extension_attributes
# Purpose: (dynamic) Get an extension attribute(s)
# Parameters: undef = Return hash of extension attributes, else return value of specific attribute
# Output: Hash or value
sub get_extension_attributes {
    my $self = assert_dynamic(shift);

    if (@_) {
	my $att = shift @_;
	my $val = $self->{_extension_hash}->{$att};
	return $val;
    }
    return %{$self->{_extension_hash}};
}

######################
# Name: get_by_extension_attribute
# Purpose: (static) Search for all pgenes that have a given extension attribute
# Parameters: A database handle, an extension attribute name, an extension value (optional)
# Output: A list of pgenes
sub get_by_extension_attribute {
    my $self = assert_static(shift);
    my $class = ref($self);
    my $dbh = assert_dbh(shift);
    my $att = shift;
    my $value = shift;

    my $where = "_att_name = ?att?".($value ? " and _value = ?val?" : "");

    my $var_name = $value ? "get_by_extension_value" : "get_by_extension";
    my $query = QUERY::Cache->get($dbh, $class, $var_name);
    unless ($query) {
	$query = create_query(type          => "SVSelect",
			      dbh          => $dbh, 
			      table        => $ext_table,
			      return_field => "_pgene_id",
			      where        => $where);
	QUERY::Cache->set($dbh, $class, "get_by_extension", $query);
    }
    $value ? $query->execute(att=>$att, val=>$value) : $query->execute(att => $att);

    return map {CLASSES::Pseudogene->get_pk($dbh, $_)} $query->fetch_values();
}

########################
# Name: get_GC_Pgene 
# Purpose: (dynamic) Return / update the GC_Pgene field
# Parameters: None
# Output: A fraction in [0,1]
sub get_GC_Pgene {
    my $self = assert_dynamic(shift);

    unless ($self->{_GC_Pgene}) {
       my $gc = $self->gc_content();
       $self->{_GC_Pgene} = $gc;
    }

    return $self->{_GC_Pgene};
}

########################
# Name: gc_content
# Purpose: (dynamic) Calculate the pgenes GC content
# Parameters: None
# Output: A fraction in [0,1]
sub gc_content {
    my $self = assert_dynamic(shift);

    my $seq = $self->get_sequence();
    my $seq2 = $seq;
    $seq2 =~ s/[^gGcC]//g;
    return (1.0*length($seq2))/length($seq);
}

##########################
# Name: get_db_extensions
# Purpose: (static) Return a list of all extensions in the database
# Parameters: None
# Output: An array
sub get_db_extensions {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    
    my $query_name = "get_extension_list";
    my $query = QUERY::Cache->get($dbh, $class, $query_name);

    unless ($query) {
	$query = create_query(type => "SVSelect",
			      dbh => $dbh,
			      table => $ext_table,
			      return_field => "_att_name",
			      command => "group by _att_name");
	QUERY::Cache->set($dbh, $class, $query_name, $query);
    }

    $query->execute();

    return $query->fetch_values();
}

#######################
# Name: get_tax_id
# Purpose: (dynamic) Return the associated tax id.
# Parameters: None
# Output: The tax. id
sub tax_id {
    my $self = assert_dynamic(shift);

    return $self->get_chromosome()->get_build()->get_tax_id();
}

########################
# Name: _add_extension_to_db
# Purpose: (dynamic) Add a new extension attribute directly to the database
# Parameters: Extension attribute, value
# Output: None
# Effects: Modifies the database without requiring a new version number
sub _add_extension_to_db {
    my $self = assert_dynamic(shift);
    my $att = assert(shift, "Need attribute name");
    my $value = assert(shift, "Need value name");

    my $class = ref($self);
    my $dbh = $self->get_dbh();

    my $query_name = "_save_insert";
    my $query = QUERY::Cache->get($dbh, $class, $query_name);

    unless ($query) {
	$query = create_query(type => "Insert",
			      dbh => $self->get_dbh(),
			      table => $ext_table,
			      attributes => ["_pgene_id", "_att_name", "_value"]);
	QUERY::Cache->set($dbh, $class, $query_name, $query);
    }

    $query->execute(_pgene_id=>$self->get_id(), _att_name=>$att, _value=>$value);
    $self->{_extension_hash}->{$att} = $value;
}

########################
# Name: get_ucsc_link
# Purpose: (dynamic) get an html link to the region on the UCSC page
# Parameters: none
# Output: html link
sub get_ucsc_link {
    my $self = assert_dynamic(shift);
    assert(@_ == 0, "Does not take parameters");
    my $tax_id = $self->tax_id();
    my $org = get_org($tax_id);
    my $clade = get_clade($tax_id);
    my $ucsc_number = $self->get_chromosome()->get_build()->get_ucsc_number();
    my $chr_number = $self->get_chromosome()->get_number();
    my ($s, $f) = $self->get_coords();

    return "" unless $clade && $ucsc_number;

    return "http://genome.ucsc.edu/cgi-bin/hgTracks?clade=$clade&org=$org&db=$ucsc_number&position=chr$chr_number\%3A$s-$f";
}
1;

