package GRAPHICS::display;
use strict;
use GD;
use POSIX("floor");
use TOOLS::assert;
require Exporter;

our @ISA = qw(Exporter);
our @EXPORT = qw(display_pgenes display_chromosome);
our $VERSION = '1.0';

######################
# Name: display_chomosome
# Purpose: Add the display of a chomosome to a map
# Parameters:
# Output: None
# sub display_chromosome {
#     my %params = @_;
#     my $map = $params{map};
#     my $chr_color = $params{chr_color} || $map->colorAllocate(0,0,0);
#     my $tic_color = $params{tic_color} || $chr_color;
#     my $tic_height = $params{tic_height} || 10;
#     my ($height, $width) = $params{dimensions} ? @{$params{dimensions}} : (($params{height}, $params{width}) || die("display_chromosome: Must provide dimensions"));
# }


######################
# Name: display_pgenes
# Purpose: Add the display of a set of pgenes to a map
# Parameters: (hash style)
#      map: The image map to which the image should be added
#      color: Reference to an array of map colors (color of chromosome bar, color of processed, color of duplicated)
#      origin: The map coordinates of the "origin"
#      dimensions: Reference to an array defining the dimensions of the area to be used (hight / width)
#      seperation: Space between rows
#      pgene_height: Height of the box for each pgene
#      tic_height: Height of the tics marking nucleotide coordinates
#      end_points: First and last nucliotde coordinate to be displayed (numbered from 1)
#      pgenes: Reference to an array of the pgenes to be planted
#      pgene_info: pgene method to be invoked for display
#      display_coords: If true, display the coordinated of each pgene in text
#      display_tics: tic seperaion in nucliotides 
#           If non-existant, a default is used
#           If false, tics not displayed      
# Output: None

sub display_pgenes {
    my %params = @_;

    my $map = $params{"map"};
    my ($chr_color, $processed_color, $duplicated_color) = @{$params{"colors"}};

    my @origin = @{$params{"origin"}};
    my ($height, $width) = @{$params{"dimensions"}};
    my $seperation = $params{"seperation"};
    my $pgene_height = $params{"pgene_height"};
    my $tic_height = $params{"tic_height"};

    my ($start_nuc, $end_nuc) = map {$_-1} @{$params{"end_points"}};
    my @pgenes = @{$params{"pgenes"}};

    
    my $num_rows = floor($height / $seperation + 1);      
    my $segment_length = $end_nuc - $start_nuc + 1;       
    my $scale_factor = ($num_rows*$width-1)/($segment_length-1);
    
    for my $pgene (@pgenes) {
        my ($start_point, $finish_point) = map {floor((($_-1) - $start_nuc)*$scale_factor)} $pgene->get_coords(); 
        my $strand = $pgene->get_strand();                                                    
	my $color = $pgene->get_ptype() eq "p" ? $processed_color : $duplicated_color;

        my $start_row = floor($start_point / $width);
        my $finish_row = floor($finish_point / $width);

	my $method = $params{"pgene_info"};
	my $str = $method ? $pgene->$method : undef;
	$str .= ": [".join("-", $pgene->get_coords())."]" if $str && $params{"display_coords"};

	if ($start_row==$finish_row) {
	    draw_pgene($map, $color, $origin[0] + $start_point - $width*$start_row, $origin[0] + $finish_point - $width*$start_row, $start_row*$seperation+$origin[1], $strand, $pgene_height, $str) if ($start_row >= 0 && $start_row < $num_rows);
	}
 	else {
	    draw_pgene($map, $color, $origin[0] + $start_point - $width*$start_row, $origin[0] + $width-1, $start_row*$seperation+$origin[1], $strand, $pgene_height, $str) if $start_row >= 0;
	    for (my $i=&max($start_row+1,0); $i < &min($finish_row, $num_rows); $i++) {
		draw_pgene($map, $color, $origin[0], $origin[0] + $width-1, $i*$seperation+$origin[1], $strand, $pgene_height);		
	    }
	    draw_pgene($map, $color, $origin[0], $origin[0] + $finish_point - $width*$finish_row, $finish_row*$seperation+$origin[1], $strand, $pgene_height, $str) if $finish_row < $num_rows;
	}
    }

    # Draw chromosome
    my $tic_seperation = !exists($params{"display_tics"}) ? 10**floor(log10($segment_length)-1) : $params{"display_tics"};
    my $current_tic = $start_nuc;
    for (my $i=0; $i < $num_rows; $i++) {
	my $y = $origin[1] + $seperation*$i;
	$map->line($origin[0], $y, $origin[0]+$width-1, $y, $chr_color);
	
	if ($tic_seperation) {
	    while ($current_tic < ($i+1)*$width/$scale_factor + $start_nuc) {
		my $tic_position = floor(($current_tic-$start_nuc)*$scale_factor - $i*$width + $origin[0]);
	    
		$map->line($tic_position, $y-$tic_height/2, $tic_position, $y+$tic_height/2, $chr_color);
		$map->string(gdSmallFont, $tic_position, $y+$tic_height/2+3, $current_tic+1, $chr_color);
		$current_tic += $tic_seperation;
	    }
	}
    }
    
    

}

sub draw_pgene {
    my $map = shift;
    my $color = shift;
    my ($x1, $x2) = (shift, shift);
    my $y = shift;
    my $strand = shift;
    my $height = shift;
    my $str = shift;

    my ($y1, $y2, $str_y) = $strand eq "+" ? ($y-$height, $y, $y+5) : ($y, $y+$height, $y-12);
    $map->filledRectangle($x1, $y1, $x2, $y2, $color);
    $map->string(gdSmallFont, $x1, $str_y, $str, $color) if $str;
}


sub min {
    return $_[0] <= $_[1] ? $_[0] : $_[1];
}  

sub max {
    return $_[0] <= $_[1] ? $_[1] : $_[0];
}

sub log10 {
    return log(shift)/log(10);
}
