package RELATIONS::Fragment_to_Pgene;
our $VERSION = '1.0';
use strict;
use CLASSES::Pseudogene;
use CLASSES::Fragment;
use CLASSES::Sequence_Alignment;
use QUERY::query_factory;
use QUERY::Where;
use TOOLS::get_date_string;
use TOOLS::assert;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(get_pgenes_from_fragment get_fragments_from_pseudogene get_pgene_alignments_from_fragment get_pgene_alignments_from_pseudogene add_pgene_fragments);

my $table_name = "Fragment_to_Pgene";

# Manages the Fragment_to_Pgene table used to map fragments to pgenes;

######################
# Name: get_pgenes_from_fragment
# Purpose: Map a fragement to its associated pgene
# Parameters: fragment object
# Output: All pgene objects associated with the given fragment
sub get_pgenes_from_fragment {
    my $frag = shift;
    assert(defined($frag) && ref($frag) eq "CLASSES::Fragment", "Bad fragment object");

    return CLASSES::Pseudogene->get("dbh"=>$frag->get_dbh(), 
				    "join_tables"=>[$table_name], 
				    "where"=>["Pseudogene.id=$table_name._Pgene_ID", "$table_name._Fragment_ID=?gpff_frag_id?"], 
				    "values"=>{"gpff_frag"=>$frag->get_id()}, 
				    "command"=>"order by acc",
				    "saved_query"=>(caller(0))[3]);
}


######################
# Name: get_fragments_from_pgene
# Purpose: Map a pgene to its associated fragements
# Parameters: pseudogene object
# Output: An array of associated fragments
sub get_fragments_from_pgene {
    my $pgene = shift;
    assert(defined($pgene) && ref($pgene) eq "CLASSES::Pseudogene", "Bad pgene object");

    return CLASSES::Fragment->get("dbh"=>$pgene->get_dbh(),
				  "join_tables"=>[$table_name],
				  "where"=>["Fragment.id=$table_name._Fragment_ID", "$table_name._Pgene_ID=?gffp_pgene_id?"],
				  "values"=>{"gffp_pgene_id"=>$pgene->get_id()},
				  "command"=>"order by $table_name._sequence_position",
				  "saved_query"=>(caller(0))[3]);
}


######################
# Name: get_pgene_alignments_from_pgene
# Purpose: Get the fragment and alignment objects for a given pgene
# Parameters: pseudogene object
# Output: An array of associated alignments
sub get_pgene_alignments_from_pgene {
    my $pgene = shift;
    assert(defined($pgene) && ref($pgene) eq "CLASSES::Pseudogene", "Bad pgene object");

    return CLASSES::Sequence_Alignment->get("dbh"=>$pgene->get_dbh(),
					    "join_tables"=>[$table_name],
					    "where"=>["Sequence_Alignment.id=$table_name._alignment", "$table_name._Pgene_ID=?gpafp_pgene_id?"],
					    "values"=>{"gpafp_pgene_id"=>$pgene->get_id()},
					    "command"=>"order by $table_name._sequence_position",
					    "saved_query"=>(caller(0))[3]);
}

#######################################
# Name: get_fragments_and_alignments_from_pgene
# Purpose: Get the fragments and alignments for a given pgene
# Parameters: pseudogene object
# Output: An array of tuple refs (fragment, alignment)
# sub get_fragments_and_alignments_from_pseudogene {
#     my $pgene = shift;
#     assert(defined($pgene) && ref($pgene) eq "CLASSES::Pseudogene", "Bad pgene object");

#     my $query = QUERY::Cache->get($pgene->get_dbh(), "Fragment_to_Pgene", (caller(0))[3]);
#     unless ($query) {
# 	$query = create_query("type"=>"Select", 
# 			      "dbh"=>$pgene->get_dbh(), 
# 			      "table_hash"=>{"Fragment_to_Pgene"=>["_Fragment_ID,_alignment"]}, 
# 			      "where"=>["Fragment_to_Pgene.Pgene_ID=?gfaafp_pgene_id?"], 
# 			      "command"=>"order by Fragment_to_Pgene.sequence_position");
# 	QUERY::Cache->set($pgene->get_dbh(), "Fragment_to_Pgene", (caller(0))[3], $query);
#     }

#     $query->execute("gfaafp_pgene_id"=>$pgene->get_id());

#     my @results;
#     while (my $ref = $query->fetch_row()) {
# 	push(@results, [$ref->{"_Fragment_ID"}, $ref->{"_alignment"}]);
#     }
#     $query->finish();

#     return map {[CLASSES::Fragment->get_pk($pgene->get_dbh(), $_->[0]), CLASSES::Sequence_Alignment->get_pk($pgene->get_dbh(), $_->[1])]} @results;
# }



########################################
# Name: add_pgene_fragments
# Purpose: Add fragments to a (previously non-existant) pseudogene
# Parameters: pseudogene object, one or more refs to a fragment/alignment pair
sub add_pgene_fragments {
    my $pgene = shift;
    assert(defined($pgene) && ref($pgene) eq "CLASSES::Pseudogene", "Bad pgene object");

    my @tuples = @_;
    assert(@tuples > 1, "Bad parameter\n");

    my $query = QUERY::Cache->get($pgene->get_dbh(), "Fragments_to_Pgene", (caller(0))[3]);
    unless ($query) {
	$query = create_query("type"=>"Insert", 
			      "dbh"=>$pgene->get_dbh(),
			      "table"=>$table_name, 
			      "attributes"=>[qw(_Fragment_ID _Pgene_ID _sequence_position _last_frag _alignment)]);
	QUERY::Cache->set($pgene->get_dbh(), "Fragments_to_Pgene", (caller(0))[3], $query);
    }

    my @tuples_ = sort {$a->[2]<=> $b->[2]} map {[$_->[0], $_->[1], $_->[0]->get_coords()]} @tuples;

    for (my $i=1; $i <= @tuples; $i++) {
	my $tuple = $tuples[$i-1];
	$query->execute("_Pgene_ID" => $pgene->get_id(), 
			"_Fragment_ID" => $tuple->[0]->get_id(),
			"_sequence_position" => $i,
			"_alignment" => $tuple->[1]->get_id(),
			"_last_frag" => ($i == @tuples ? 'T' : 'F'));
    }
}


###############################################
# Name: remove_pgene_from_mapping
# Purpose: Remove a pgene (and its associated mappings)
# Parameters: pseudogene object
# Output: None
sub remove_pgene_from_mapping {
    my $pgene = shift;
    assert (defined($pgene) && ref($pgene) eq "CLASSES::Pseudogene", "Bad pgene object");

    my $query = QUERY::Cache->get($pgene->get_dbh(), "Fragment_to_Pgene", "remove_pgene_from_mapping");
    unless ($query) {
	$query = create_query("type"=>"Remove", 
			      "dbh"=>$pgene->get_dbh(),
			      "table"=>$table_name, 
			      "where"=>"$table_name.Pgene_ID=?rpgm_id?");
	QUERY::Cache->set($pgene->get_dbh(), "Fragment_to_Pgene", "remove_pgene_from_mapping", $query);
    }

    $query->execute("rpgm_id"=>$pgene->get_id());
}
