package RELATIONS::Set_Mapping;
our $VERSION = '1.0';
use strict;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(set_has get_fragments get_pgenes get_set_objects get_sets_containing set_insertion);

use CLASSES::PFSet;
use CLASSES::Pseudogene;
use CLASSES::Fragment;

my @frag_table = ("PFSet_to_Fragment", "PFSet_ID", "Fragment_ID");
my @pgene_table = ("PFSet_to_Pgene", "PFSet_ID", "PGene_ID");

#######################
# Name: object_type
# Purpose: Select the type of object
# Pramaters: An object (pgene or fragment), calling method (for debugging)
# Output: 0 for fragment, 1 for pseudogene;
sub table_info {
    my $obj = shift;
    my $caller = shift;
    return 0 if ref($obj) eq "CLASSES::Fragment";
    return 1 if ref($obj) eq "CLASSES::Pseudogene";
    die("RELATIONS::Set_Mapping::$caller: Object not a fragment or pseudogene");
}

#######################
# Name: check_db
# Purpose: Check that two objects have the same db handle
# Parameters: Two objects, a caller
# Output: 1 (calls an error if false)
sub check_db {
    my ($obj1, $obj2, $caller) = @_;
    return 1 if ($obj1->get_dbh() == $obj2->get_dbh());
    die("RELATIONS::Set_Mapping::$caller: Objects have conflicting database handles\n");
}

##################################
# Name: set_has
# Purpose: Does a set have a given object (pgene or fragment)
# Parameters: One set object, one pgene object
# Output: 0/1
sub set_has {
    my $pfset = shift;
    my $obj = shift;

    check_db($pfset, $obj, "set_has");
    my @table_info = (table_info($obj, "set_has") ? @pgene_table : @frag_table);

    my @ids = ($pfset->get_id(), $obj->get_id());
    my $query = "select * from $table_info[0] where $table_info[1]=$ids[0] and $table_info[2]=$ids[1]";
    my $sth = $pfset->get_dbh()->prepare($query);
    $sth->execute();
    my $ref = $sth->fetchrow_arrayref();
    $sth->finish();

    return defined($ref);
}


##################################
# Name: get_fragment_ids
# Purpose: Get all fragment ids for a pfset
# Parameters: A pfset
# Output: A list of fragment ids
sub get_fragment_ids {
    my $pfset = shift;

    my $query = "select $frag_table[2] from $frag_table[0] where $frag_table[1]=".$pfset->get_id();
    my $sth = $pfset->get_dbh()->prepare($query);
    $sth->execute();
    
    my @frag_ids;
    while (my $ref = $sth->fetchrow_arrayref()) {
	push(@frag_ids, $ref->[0]);
    }
    $sth->finish();
    
    return @frag_ids;
}

##################################
# Name: get_fragments
# Purpose: Get all fragment ids for a pfset
# Parameters: A pfset 
# Output: A list of fragment objects
sub get_fragments {
    return map {CLASSES::Fragment->get_pk($_[0]->get_dbh(), $_)} get_fragment_ids($_[0]);
}

##################################
# Name: get_pgene_ids
# Purpose: Get all fragment ids for a pfset
# Parameters: A pfset
# Output: A list of pgene ids
sub get_pgene_ids {
    my $pfset = shift;

    my $query = "select $pgene_table[2] from $pgene_table[0] where $pgene_table[1]=".$pfset->get_id();
    my $sth = $pfset->get_dbh()->prepare($query);
    $sth->execute();
    
    my @pgene_ids;
    while (my $ref = $sth->fetchrow_arrayref()) {
	push(@pgene_ids, defined($ref) ? $ref->[0] : undef);
    }
    $sth->finish();
    
    return @pgene_ids;
}
   

##################################
# Name: get_pgenes
# Purpose: Get all fragment ids for a pfset
# Parameters: A pfset
# Output: A list of fragment objects
sub get_pgenes 
{
    return map {CLASSES::Fragment->get_pk($_[0]->get_dbh(), $_)} get_pgene_ids($_[0]);
}

##################################
# Name: get_set_objects
# Purpose: Get all pgene and fragment objects in the set
# Parameters: A pfset
#             A boolean indicating if they should be split (default false)
# Ouput: If not split: a list of fragment and pgene objects
#        If split: Two refs -- one to a list of fragment objects, one to
#                  to a list of pseudogene objects
sub get_set_objects {
    my ($pfset, $split) = @_;
    my @frags = get_fragments($pfset);
    my @pgenes = get_pgenes($pfset);

    return $split ? ([@frags], [@pgenes]) : (@frags, @pgenes);
}

#################################
# Name: get_set_ids_containing
# Purpose: Get all sets containing any of the given objects
# Parameters: One or more objects (each a pgene or fragment)
# Output: A list of all set ids containing the objects
sub get_set_ids_containing {
    return undef unless @_ > 0;

    my @objects = @_ ;
    my $ref_object = $objects[0];

    my @fq_arr;
    my @pq_arr;
    foreach (@objects) {
	check_db($ref_object, $_, "get_set_ids_containing");
	if (ref($_) eq "CLASSES::Fragment") {
	    push(@fq_arr, "$frag_table[2]=".$_->get_id());
	}
	elsif (ref($_) eq "CLASSES::Pseudogene") {
	    push(@pq_arr, "$pgene_table[2]=".$_->get_id());
	}
	else {
	    die("Set_Mapping::get_set_ids_containing: Bad object (not a fragment of pgene): $_\n");
	}
    }


    my $fquery = (@fq_arr > 0) ? "select $frag_table[0].$frag_table[1] from $frag_table[0] where (".join(" or ", @fq_arr).")" : undef;
    my $pquery = (@pq_arr > 0) ? "select $pgene_table[0].$pgene_table[1] from $pgene_table[0] where (".join(" or ", @pq_arr).")" : undef;

    my $query = (defined($fquery) ? $fquery : "").(defined($fquery) && defined($pquery) ? " union " : "").(defined($pquery) ? $pquery : "");
    print "$query\n";
    
    my $sth = $ref_object->get_dbh()->prepare($query);
    $sth->execute();
    
    my @set_ids;
    while (my $ref = $sth->fetchrow_arrayref()) {
 	push(@set_ids, $ref->[0]);
    }
    
    return @set_ids;
}	

#################################
# Name: get_sets_containing
# Purpose: Get all sets containing any given objects
# Parameters: One or more objects (each a pgene or fragment)
# Output: A list of all set objects containing the objects
sub get_sets_containing {
    return undef unless @_ > 0;
    return map {CLASSES::PFSet->get_pk($_[0]->get_dbh(), $_)} get_set_ids_containing(@_);
}


##################################
# Name: set_insertion
# Purpose: Add fragments to a pfset
# Parameters: A pfset object 
#             One or more objects (pgenes or fragments)
# Output: None
sub set_insertion {
    my $pfset = shift;
    my @objs = @_;

    my $frag_query = "insert into $frag_table[0]($frag_table[1], $frag_table[2]) values(".$pfset->get_id().", ?)";
    my $pgene_query = "insert into $pgene_table[0]($pgene_table[1], $pgene_table[2]) values(".$pfset->get_id().", ?)";
    my $f_sth = $pfset->get_dbh()->prepare($frag_query);
    my $p_sth = $pfset->get_dbh()->prepare($pgene_query);
    foreach (@objs) {
	(set_has($pfset, $_)) and die("REALTIONS::Set_Mapping::set_insertion: Attempted to insert a fragment already in a set\n");

	check_db($pfset, $_, "set_insertion");
	my $sth = table_info($_, "set_insertion") ? $p_sth : $f_sth;
	$sth->bind_param(1, $_->get_id());
	$sth->execute();
    }
}
