package SEARCH::Search;
our $VERSION = '1.0';
use strict;

use TOOLS::assert;
use CLASSES::Pseudogene;
use CLASSES::PFSet;
use CLASSES::Protein;
use QUERY::Cache;
use QUERY::query_factory;

my $get_components = 1;

my @fields = qw(name search_type search_method relations sample_generate value_type box_length value_table value_field);
my %extension_fields = (search_type => "Pseudogene", search_method => "search_extended", relations => "eq,=|ne,!=|lt,<|gt,>"); 

#######################
# Name: new
# Purpose: Constructor
# Parameters: dbh
# Output: The new object
sub new {
   my $class = shift;
 
   my $dbh = assert_dbh(shift);
   my $H = {"_dbh" => $dbh, "_sample_pgene" => (CLASSES::Pseudogene->get(dbh=>$dbh, command=>"limit 1"))[0]};

   my $self = bless $H, $class;
   return $self;
}

######################
# Name: get_dbh
# Purpose: get the dbh object
# Parameters: None
# Output the dbh object
sub get_dbh {
   my $self = assert_dynamic(shift);
   return $self->{_dbh};
}

###################################
# Funtions to get sample inputs
sub get_location() {
    my $self = assert_dynamic(shift);
    my ($s, $f) = $self->{_sample_pgene}->get_coords();
    my $chr = $self->{_sample_pgene}->get_chromosome()->get_number();

    return "chr$chr:".join("-", ($s-1, $f+1));
}

sub get_chromosome() {
    my $self = assert_dynamic(shift);

    return $self->{_sample_pgene}->get_chromosome()->get_number();
}

sub cmp_chr {
    my $isNumA = ($a =~ /^\d+/);
    my $isNumB = ($b =~ /^\d+/);

    return $isNumB ? $a <=> $b : -1 if $isNumA;
    return 1 if $isNumB;
    return $a cmp $b;
}


sub all_chromosomes {
    my $self = assert_dynamic(shift);

    my @arr = $self->{_sample_pgene}->get_chromosome()->get_build()->chromosome_numbers();
    return  sort cmp_chr @arr;
}

sub get_protein {
    my $self = assert_dynamic(shift);
    my $P = $self->{_sample_pgene}->get_protein();
    return $P->get_name() || $P->get_acc() if $P;

    my $fragment_relation_table = TOOLS::constants::FRAGMENT_RELATION_TABLE;
    my $dbh = $self->get_dbh();
    my @join_tables = ($fragment_relation_table, "Sequence_Alignment", "Protein");
    my @where = ("Pseudogene.id=$fragment_relation_table._pgene_id", "$fragment_relation_table._alignment=Sequence_Alignment.id", "Sequence_Alignment.protein_id is not NULL");

    my $class = ref($self);
    $P = CLASSES::Pseudogene->get("dbh"=>$dbh,
		     "join_tables"=>\@join_tables,
		     "where"=>\@where,
		     "command"=>"limit 1",
		     "saved_query"=>"get_pgene_with_known_protein", @_);
    return $P ? P->get_name() || $P->get_acc() : "No known proteins";
}

sub get_gene {
    my $self = assert_dynamic(shift);
    my $P = $self->{_sample_pgene}->get_protein();
    return $P->get_gene_id() if $P && $P->get_gene_id();

    my $fragment_relation_table = TOOLS::constants::FRAGMENT_RELATION_TABLE;
    my $dbh = $self->get_dbh();
    my @join_tables = ($fragment_relation_table, "Sequence_Alignment", "Protein");
    my @where = ("Pseudogene.id=$fragment_relation_table._pgene_id", "$fragment_relation_table._alignment=Sequence_Alignment.id", "Sequence_Alignment.protein_id=Protein.id", "Protein.gene_id is not NULL");

    my $class = ref($self);
    $P = CLASSES::Pseudogene->get("dbh"=>$dbh,
		     "join_tables"=>\@join_tables,
		     "where"=>\@where,
		     "command"=>"limit 1",
		     "saved_query"=>"get_pgene_with_known_protein", @_);
    return $P ? $P->get_protein()->get_gene_id() : "No known genes";
}

sub get_type {
    my $self = assert_dynamic(shift);
    return $self->{_sample_pgene}->get_ptype();
}

sub get_source {
    my $self = assert_dynamic(shift);
    return $self->get_source();
}

sub get_gc_content {
    my $self = assert_dynamic(shift);
    my $P = $self->{_sample_pgene};
    my $gc = $P->get_GC_Pgene();
    my $s = $gc < .06 ? $gc/2 : $gc-0.05;
    my $f = $gc > .94 ? $gc + (1-$gc/2) : $gc+0.05;

    my $str = sprintf("%0.2f-%0.2f", $s, $f);
    return $str;
}

sub get_acc {
    my $self = assert_dynamic(shift);
    return $self->{_sample_pgene}->get_acc();
}

sub get_id {
    my $self = assert_dynamic(shift);
    return $self->{_sample_pgene}->get_id();
}

sub get_blank {
    return "";
}
    
###################################
# Search functions

######################
# Name: pgene_by_acc
# Purpose: Returns all pseudogenes in the acc range
# Parameters: 
# Output:  All psedugenes overlapping the given acc
sub pgene_by_acc {
    my $self = assert_dynamic(shift);
    my $acc_str = assert(shift, "Must provide an accession number");
    $acc_str =~ /(\d+)$/;
    my $acc = $1;

    my $operator = assert(shift, "No operator provided");
    my $extended_name = shift;  # Should be undefined
    my $filter_str = shift;
    my $versions = shift;

    return CLASSES::Pseudogene->get(dbh=>$self->get_dbh(), where=>"acc $operator $acc", "where_filter"=>$filter_str, all_versions=>$versions, get_components=>$get_components);
}

######################
# Name: pgene_by_id
# Purpose: Returns all pseudogenes in the acc range
# Parameters: 
# Output:  All psedugenes overlapping the given id
sub pgene_by_id {
    my $self = assert_dynamic(shift);
    my $id = assert(shift, "Must provide an accession number");

    my $operator = assert(shift, "No operator provided");
    my $extended_name = shift;  # Should be undefined
    my $filter_str = shift;
    my $versions = shift;

    return CLASSES::Pseudogene->get(dbh=>$self->get_dbh(), where=>"id $operator $id", "where_filter"=>$filter_str, all_versions=>$versions, get_components=>$get_components);
}

######################
# Name: pgene_by_gene_id
# Purpose: Returns all pseudogenes corresponding to the given genes
# Parameters: 
# Output:  All psedugenes overlapping the given gene id
sub pgene_by_gene_id {
    my $self = assert_dynamic(shift);
    my $gene_id = assert(shift, "Must provide a gene id");

    my $operator = assert(shift, "No operator provided");
    my $extended_name = shift;  # Should be undefined
    my $filter_str = shift;
    my $versions = shift;

    $gene_id =~ s/^\s+([^\s])\s+$/$1/;

    my $join_tables = ["Protein", "Sequence_Alignment", "Fragment_to_Pgene"];
    my $where = QUERY::Where->new(clauses=>["Protein.gene_id='$gene_id'",
					    "Protein.id=Sequence_Alignment.protein_id",
					    "Sequence_Alignment.frag_id=Fragment_to_Pgene._Fragment_ID",
					    "Fragment_to_Pgene._Pgene_ID=Pseudogene.id"]);

    return CLASSES::Pseudogene->get(dbh=>$self->get_dbh(), join_tables=>$join_tables, where=>$where, where_filter=>$filter_str, all_versions=>$versions, get_components=>$get_components);
#    return CLASSES::Pseudogene->get_by_gene($self->get_dbh(), $gene_id, "where_filter" => $filter_str, all_versions=>$versions, get_components=>$get_components);
}


######################
# Name: pgene_by_location
# Purpose: Returns all pseudogenes overlapping a given area 
# Parameters: location string (chrX:100-200), build (default = current)
# Output:  All psedugenes overlapping the given location
sub pgene_by_location {
   my $self = assert_dynamic(shift);
   my $location_str = assert(shift, "Must provide a location");
   my $operator = assert(shift, "No operator provided");
   my $extended_name = shift;  # Should be undefined
   my $filter_str = shift;
   my $versions = shift;

   $location_str =~ s/\s//g;
   unless ($location_str =~ /^(chr){0,1}(\w+)(:(\d+)-(\d+)){0,1}$/) {
       assert(0==1, "Bad location string: $location_str.  Must be of the form '22', 'chr22', or 'chr22:1000-2000'");
   }
   my ($chr, $start, $stop) = ($2, $4, $5);

   my $join_tables = ["Chromosome", "Fragment_to_Pgene as Start", "Fragment as StartFragment", "Chromosome_Segment as StartSeg", "Fragment_to_Pgene as Finish", "Fragment as FinishFragment", "Chromosome_Segment as FinishSeg"];
   
   my $chromosome_number = "Chromosome.number = '$chr'";                      
   my $choose_segments = "StartSeg.chr_id = Chromosome.id and FinishSeg.chr_id = Chromosome.id";
   my $fragment_endpoint = "StartSeg.id=StartFragment.start_seg and FinishSeg.id=FinishFragment.finish_seg";
   my $start_fragment =  "Pseudogene.id=Start._Pgene_ID and Start._sequence_position=1 and Start._Fragment_ID=StartFragment.id";
   my $finish_fragment  = "Pseudogene.id=Finish._Pgene_ID and Finish._last_frag='T' and Finish._Fragment_ID=FinishFragment.id";
   my $frag_coords = $operator eq "overlap" ? "StartSeg.start+StartFragment.start_offset <= $stop and FinishSeg.start+FinishFragment.finish_offset >= $start" :
                     $operator eq "contains" ? "StartSeg.start+StartFragment.start_offset >= $start and FinishSeg.start+FinishFragment.finish_offset <= $stop" :
		     $operator eq "contained by" ? "StartSeg.start+StartFragment.start_offset <= $start and FinishSeg.start+FinishFragment.finish_offset >= $stop" :
		     die("Bad operator");
   my $where = QUERY::Where->new(clauses => [$chromosome_number, $choose_segments, $fragment_endpoint, $start_fragment, $finish_fragment, $frag_coords]);

   return CLASSES::Pseudogene->get(dbh=>$self->get_dbh(), join_tables=>$join_tables, where=>$where, where_filter=>$filter_str, all_versions=>$versions, get_components=>$get_components);
}

######################
# Name: pgene_by_chromosome
# Purpose: Returns all pseudogenes on a given chromosome
# Parameters: chromsome number, build (default = current)
# Output:  All psedugenes on the given chromosome
sub pgene_by_chromosome {
   my $self = assert_dynamic(shift);
   my $chr = assert(shift, "Must provide a chromosome");
   my $operator = assert(shift, "No operator provided");
   my $extended_name = shift;  # Should be undefined
   my $filter_str = shift;
   my $versions = shift;

   my $join_tables = ["Fragment_to_Pgene", "Fragment", "Chromosome_Segment", "Chromosome"];
   my  $where = QUERY::Where->new(clauses=>["Chromosome.number = '$chr'",
					    "Chromosome.id = Chromosome_Segment.chr_id",
					    "Chromosome_Segment.id = Fragment.start_seg",
					    "Fragment.id = Fragment_to_Pgene._Fragment_ID",
					    "Fragment_to_Pgene._Pgene_ID = Pseudogene.id"]);

   return CLASSES::Pseudogene->get(dbh=>$self->get_dbh(), join_tables => $join_tables, where_filter=>$filter_str, where=>$where, all_versions=>$versions, get_components=>$get_components);
#   return  CLASSES::Pseudogene->get_by_location($self->get_dbh(), $C, undef, undef, "where_filter"=>$filter_str, all_versions=>$versions, get_components=>$get_components, "debug-print"=>1);
}

######################
# Name: pgene_by_protein
# Purpose: Return all pseudogenes using a given area
# Parameters: protein value, merge string
# Output: All pseudogenes corresponding to the protein acc/name
sub pgene_by_protein {
    my $self = assert_dynamic(shift);
    my $protein_str = assert(shift, "No protein id provided");
    my $operator = shift;
    my $extended_name = shift;  # Should be undefined
    my $filter_str = shift;
    my $versions = shift;

    $protein_str =~ s/^\s+([^\s])\s+$/$1/;

    my $dbh = $self->get_dbh();
    my $primary = CLASSES::Protein->get_primary_acc($dbh, $protein_str);
    my $protein = $primary ? CLASSES::Protein->get_by_acc($dbh, $primary) : CLASSES::Protein->get(dbh=>$dbh, where=>"name = '$protein_str'");
    
    return  $protein ? CLASSES::Pseudogene->get_by_protein($dbh, $protein->get_acc(), "where_filter" => $filter_str, all_versions=>$versions, get_components=>$get_components) : ([], QUERY::Where->new(), {});
}

######################
# Name: pgene_by_gc
# Purpose: Return all pseudogenes with a gc content within a given range
# Parameters: gc contant range (e.g. 0.4-0.6), merge string
# Output: All pseudogenes with a gc-content within the range
sub pgene_by_gc {
    my $self = assert_dynamic(shift);
    my $range_str = assert(shift, "No gc range provided");
    my $operator = assert(shift, "No operator provided");
    my $exntended_name = shift;  # Should be undefined
    my $filter_str = shift;
    my $versions = shift;

    my $dbh = $self->get_dbh();

    my ($lower, $upper);
    if ($range_str =~ /^(0.*\d*)-(0.*\d*)$/) {
	($lower, $upper) = ($1, $2);
    }
    else {
	assert(1==2, "Bad gc range string: $_");
    }

    return CLASSES::Pseudogene->get(dbh=>$dbh, where=>"GC_Pgene >= $lower and GC_Pgene <= $upper", "where_filter" => $filter_str, all_versions=>$versions, get_components=>$get_components);
}

######################
# Name: pgene_by_type
# Purpose: Return all pseudogenes of a given type
# Parameters: Standard
# Output: 
sub pgene_by_type {
    my $self = assert_dynamic(shift);
    my $type = assert(shift, "No type name provided");
    my $operator = assert(shift, "No operator provided");
    my $extended_name = shift;  # Should be undefined
    my $filter_str = shift;
    my $versions = shift;

    return CLASSES::Pseudogene->get(dbh=>$self->get_dbh(), where=>"ptype $operator '$type' ", "where_filter" => $filter_str, all_versions=>$versions, get_components=>$get_components);
}

######################
# Name: pgene_by_source
# Purpose: Return all pseudogenes found by a given source
# Parameters: source name, operator, filter string
sub pgene_by_source {
    my $self = assert_dynamic(shift);
    my $source = assert(shift, "No source name provided");
    my $operator = assert(shift, "No operator provided");
    my $extended_name = shift;  # Should be undefined
    my $filter_str = shift;
    my $versions = shift;

    if ($operator eq 'like') {
	$source = "\%$source\%";
    }

    return CLASSES::Pseudogene->get(dbh=>$self->get_dbh(), where=>"(source $operator '$source') ", "where_filter" => $filter_str, all_versions=>$versions, get_components=>$get_components);
}

sub set_by_name {
    my $self = assert_dynamic(shift);
    my $name = assert(shift, "No set name provided");
    my $operator = assert(shift, "No operator provided");
    my $extended_name = shift;
    my $filter_str = shift;
    my $versions = shift;

    $name = "\%$name\%" if $operator eq 'like';
    return CLASSES::PFSet->get(dbh=>$self->get_dbh(), where=>"name $operator '$name'", where_filter=>$filter_str, all_versions=>$versions);
}

sub set_by_notes {
    my $self = assert_dynamic(shift);
    my $notes = assert(shift, "No set name provided");
    my $operator = assert(shift, "No operator provided");
    my $extended_name = shift;
    my $filter_str = shift;
    my $versions = shift;


    $notes = "\%$notes\%";
    return CLASSES::PFSet->get(dbh=>$self->get_dbh(), where=>"notes $operator '$notes'", where_filter=>$filter_str, all_versions=>$versions);
}

sub set_by_acc {
    my $self = assert_dynamic(shift);
    my $acc = assert(shift, "No accession number provided");
    my $operator = assert(shift, "No operator provided");
    my $extended_name = shift;
    my $filter_str = shift;
    my $versions = shift;


    return CLASSES::PFSet->get(dbh=>$self->get_dbh(), where=>"PFSet.acc $operator $acc", where_filter => $filter_str, all_versions=>$versions);
}

sub set_by_id {
    my $self = assert_dynamic(shift);
    my $id = assert(shift, "No id provided");
    my $operator = assert(shift, "No operator provided");
    my $extended_name = shift;
    my $filter_str = shift;
    my $versions = shift;

    return CLASSES::PFSet->get(dbh=>$self->get_dbh(), where=>"PFSet.id $operator $id", where_filter => $filter_str, all_versions=>$versions);
}


######################
# Name: search_extended
# Purpose: Return all pseudogenes where the given extended attriute relates to its value as specified
# Parameters: value, mysql relation operator, extention type, merge string
# Output: All pseudogenes found by the search
sub search_extended {
    my $self = assert_dynamic(shift);   my $class = ref($self);
    my $value = shift;  assert(defined $value, "No value provided");
    my $operator = assert(shift, "No operator provided");
    my $type = assert(shift, "No type provided");
    my $filter_str = shift;
    my $all_versions = shift;

    my $join_tables = ["Pseudogene_Extensions"];
    my $where = QUERY::Where->new(clauses=>["Pseudogene_Extensions._att_name = '$type'", 
					    "Pseudogene_Extensions._value $operator '$value'", 
					    "Pseudogene_Extensions._Pgene_id = Pseudogene.id"]);

    return CLASSES::Pseudogene->get(dbh=>$self->get_dbh(), join_tables=>$join_tables, where=>$where, all_versions=>$all_versions, whre_filter=>$filter_str, get_components=>$get_components);
}
 
######################
# Name: search
# Purpose: Invoke 1 search method
# Parameters: method, value, operator, "limiting values"
# Output: A list of the results
sub search {
    my $self = assert_dynamic(shift);
    my $ext_name = undef;
    my ($method, $value, $operator, $filter_str, $versions, $join_tables) = (shift, shift, shift, shift, shift);
    if ($method =~ /(search\_extended)\_(.+)/) {
	($method, $ext_name) = ($1, $2);
    }

    return $self->$method($value, $operator, $ext_name, $filter_str, $versions, $join_tables);
}

#######################
# Name: extract_searches
# Purpose: (dynamic) Extract search information for presentation on the web page
# Parameters: Type being searched for (Pseudogene or PFSet)
# Output: An arrray of array refs., each of the form [name, type, method, relations, sample_generate, value_type, length | (value_table, value_file)] 
sub extract_searches {
   my $self = assert_dynamic(shift);
   my $class = ref($self);
   my $dbh = $self->get_dbh();
   my $search_type = shift;

   my $query_name = "extract_searches_query";
   my $query = QUERY::Cache->get($dbh, $class, $query_name);

   unless ($query) {
      my $table_name = "Search";
      $query = create_query(type => "Select",
	                   dbh => $dbh,
			   table_hash => {$table_name => \@fields},
			   where => "search_type = '$search_type'");
      QUERY::Cache->set($dbh, $class, $query_name, $query);
   }
   $query->execute();

   my @results;
   for ($query->fetch_all()) {
      my %H = %$_;
      my @values = map {$H{$_}} @fields;
      push(@results, \@values); #[$H{name}, $H{search_type}, $H{search_method}, $H{relations}, $H{sample_value}]);
   }

   if ($search_type eq 'Pseudogene') {
       $query_name = "extract_extended_names_query";
       $query = QUERY::Cache->get($dbh, $class, $query_name);
       
       unless ($query) {
	   my $table_name = "Pseudogene_Extensions";
	   my @extended_fields = qw(_att_name _value); 
	   $query = create_query(type => "Select", dbh => $dbh, table_hash => {$table_name => \@extended_fields}, command => "group by _att_name");
       }
       $query->execute();
       
       for ($query->fetch_all()) {
	   my %H = %$_;
	   $extension_fields{name} = $H{_att_name};
	   $extension_fields{sample_generate} = $H{_value};
	   $extension_fields{search_method} = "search_extended_".$H{_att_name};
	   my @values = map {$extension_fields{$_}} @fields;
	   push(@results, \@values);
	   #push(@results, [$H{"_att_name"}, "Pseudogene", $extended_method, $extended_relations, $H{"_value"}]);
       }
   }

   return @results;
}

1;
