#!/usr/bin/perl -w
use strict;

use CLASSES::ClassFactory;
use CLASSES::Fragment;
use CLASSES::Sequence_Alignment;
use CLASSES::Pseudogene;
use CLASSES::Build;
use ALIGNMENT::default_align;
use DB::db;

#######################################################################
# remap: Remap the fragments of one build to another
# Currently assumes that both builds are in the database, and that
# the source build is the most current.  (e.g. All current objects
# are with respect to that build.)
# Upon completion, all current objects will be with respect to
# the target build.
#
# For future: should probably have it add in the new build, instead of
# assuming it exists.

######################
# Input constants
my $source_build_number = $ARGV[0] || "34"; #die("Need a source build\n");
my $target_build_number = $ARGV[1] || "35"; #die("Need a target build\n");
my $liftOver_file = $ARGV[2] || "/data/home/karro/tools/liftOver/hg16ToHg17.over.chain"; #die("Need a liftover file\n"); #"/data/home/karro/tools/liftOver/hg10ToHg16.over.chain";
my $tax_id = $ARGV[3] || "9606"; #die("Need a tax id\n"); #9606;

######################
# Database initilization
my $dbh = DB::db::getDB("db" => "Pgenes_$tax_id");

######################
# Other constants
my $source_build = CLASSES::Build->get_by_number($dbh, $source_build_number) || die("Source build not found in database\n");  # Source build object
my $target_build = CLASSES::Build->get_by_number($dbh, $target_build_number) || die("Target build not found in database\n");  # Target build object
my $source_file = TOOLS::constants::SOURCE_TEMP;       # Input file for liftOver execution
my $target_file = TOOLS::constants::TARGET_TEMP;       # Output file for liftOver execution
my $unmapped_file = TOOLS::constants::UNMAPPED_TEMP;   # Unmapped output file for liftOver execution
my $liftOver = TOOLS::constants::LIFTOVER;             # liftOver executable
my $comment_file = "comments.txt";                     # Write execution notes
open(COMMENT_FILE, ">$comment_file");

######################
# Data structure
my %source_build_chromosomes;                          # Hash of chromosome number to source build chromosomes
my %target_build_chromosomes;                          # Hash of chromosome number of tartget build chromosomes
my @old_fragments;                                     # List of all current fragments in the database
my %old_fragment_hash;                                 # Hash of old id to old frargment.
my %acc_hash;                                          # Map accs to old ids.
my %new_fragment_hash;                                 # Hash of old id to new fragment
my %alignment_hash;                                    # Mapping of old fragment id to a list of its alignments
my %alignment_mapping;                                 # Map old alignments to their new versions
my @old_pgenes;                                        # List of all current pgenes
my %pgene_hash;                                        # Mapping of old fragement id to its list of pgenes                                   
my @sorted_new_fragments;                              # List of new fragments, sorted by location
my @conflict_intervals;                                # All fragments in @sorted_new_fragments[$conflict_intervals[$i], $confict_intervales[$i+1]-1] are the same

######################
# Create data structures
%source_build_chromosomes = map {($_->get_number() => $_)} $source_build->get_chromosome();
%target_build_chromosomes = map {($_->get_number() => $_)} $target_build->get_chromosome();

my $limit = undef; #"limit 1000";
@old_fragments = CLASSES::Fragment->get("dbh"=>$dbh, command=>$limit); #,  "where"=>"(Fragment.id = 13077)");
@old_pgenes = CLASSES::Pseudogene->get("dbh"=>$dbh); #, where=>"id=48850", command=>$limit); #, "where"=>"(Pseudogene.id = 13077)");
#@old_fragments = map {$_->get_fragments()} @old_pgenes;
print("Size: ".@old_fragments." ".@old_pgenes."\n");
for my $frag (@old_fragments) {
    my $id = $frag->get_id();
    my @alignments = CLASSES::Sequence_Alignment->get_by_fragment($dbh, $frag);
    my @pgenes = CLASSES::Pseudogene->get_by_fragment($dbh, $frag);

    $acc_hash{$frag->get_acc()} = $id;
    $old_fragment_hash{$id} = $frag;
    $new_fragment_hash{$id} = $frag->new();        # Still need to set the new location
    $alignment_hash{$id} = \@alignments;
    $pgene_hash{$id} = \@pgenes;
}

######################
# Execute liftOver and update fragments
print "Step: liftOver\n";
open(OFILE, ">$source_file") || die("Can't open source file: $source_file\n");
for (@old_fragments) {
   print OFILE $_->gtf_string()."\n";
}
close(OFILE);

system("$liftOver -gff $source_file $liftOver_file $target_file $unmapped_file");

open(IFILE, "$unmapped_file");
while (<IFILE>) {      # Eliminate fragments that cannot be remapped
    <IFILE> =~ /id (\w+)$/ || die("id not found: $_\n");
    my $id = $1;
    my $acc = $old_fragment_hash{$id}->get_acc();
    $new_fragment_hash{$id} = undef;
    print COMMENT_FILE "Fragment not remapped: $id\n";
    CLASSES::Fragment->mark_dead($dbh, $acc);
}
close(IFILE);

open(IFILE, "$target_file");
while (<IFILE>) {
    /^chr(\w+)\t.+\t.+\t(\d+)\t(\d+)\t.+\t(\+|\-)\t.+id (\w+)$/ || die("Bad line: $_");  # chr, start, finish, strand, id	
    my ($chr, $start, $finish, $strand, $id) = ($1, $2, $3, $4, $5);
    if ($target_build_chromosomes{$chr}) {
        $new_fragment_hash{$id}->set_location($target_build_chromosomes{$chr}, $start, $finish, $strand);
    }
    else { # Mapped to a non-exsitant / unrecorded choromosome
	my $acc = $old_fragment_hash{$id}->get_acc();
	$new_fragment_hash{$id} = undef;
	print COMMENT_FILE "Bad chromosome: $id, $chr\n";
	CLASSES::Fragment->mark_dead($dbh,$acc);
    }
}
close(IFILE);

#######################
# Set up @sorted_new_fragments and @conflict_intervals
print "Step: Set up\n";
@sorted_new_fragments = sort {$a->get_chromosome->get_number() cmp $b->get_chromosome->get_number() || ($a->get_coords())[0] <=> ($b->get_coords())[0] || ($a->get_coords())[1] <=> ($b->get_coords())[1] || $a->get_strand() cmp $b->get_strand() || $a->get_acc() <=> $b->get_acc()} grep {$_} values %new_fragment_hash;

my $counter = 0;
while ($counter < @sorted_new_fragments) {
    push(@conflict_intervals, $counter);
    my $next = $counter+1;
    $next++ while ($next < @sorted_new_fragments && compare_fragments($sorted_new_fragments[$counter], $sorted_new_fragments[$next]));
    $counter = $next;
}
push(@conflict_intervals, scalar @sorted_new_fragments);

#######################
# Insert new fragments
print "Step: Insert\n";
for (my $i=0; $i < @conflict_intervals-1; $i++) {
    my $counter = $conflict_intervals[$i];
    my $last = $conflict_intervals[$i+1]-1;
    my $base_new_frag = $sorted_new_fragments[$counter];
    $base_new_frag->insert();
    for (@sorted_new_fragments[($counter+1)..$last]) {	
	my $old_id = $acc_hash{$_->get_acc()};
        $new_fragment_hash{$old_id} = undef;      # Use below line to retain and update sequence alignment for conflicting fragments
	# $new_fragment_hash{$old_id} = $base_new_frag;
	CLASSES::Fragment->mark_dead($dbh, $_->get_acc());

	print COMMENT_FILE "Fragment $old_id not updated due to remapping conflict\n";
    }
}
	
	  

#######################
# Update seqeunce alignments
print "Step: Alignments\n";
while (my ($old_id, $new_fragment) = each %new_fragment_hash) {
    next unless $new_fragment;
    my $old_fragment = $old_fragment_hash{$old_id};
    #my $changed = $old_fragment->get_sequence() ne $new_fragment->get_sequence();
    for my $old_align (CLASSES::Sequence_Alignment->get_by_fragment($dbh, $old_fragment)) {
	my $new_align;
	my $protein = $old_align->get_protein();
	$new_align = default_align($protein, $new_fragment);
	if (defined $new_align && ref($new_align) eq "CLASSES::Sequence_Alignment") {
	    $new_align->insert();
	    $alignment_mapping{$old_align->get_id()} = $new_align;
	}
	else {
	    #$new_fragment_hash{$old_id} = undef;                            # ??? These two lines must be a mistake
	    #CLASSES::Fragment->mark_dead($dbh, $new_fragment->get_acc());
            $alignment_mapping{$old_align->get_id()} = undef;
	    print COMMENT_FILE "Unable to re-align fragment ".$new_fragment->get_id()."(from fragment $old_id)\n";
	}
    }
}

#######################
# Update pseudogenes 
print "Step: Pseudogenes\n";
# Assumes conflicts were not updated, and requries that all fragments of a pgene be updated
for my $old_pgene (@old_pgenes) {
    my @old_fragments = $old_pgene->get_fragments();
    my @old_alignments = $old_pgene->get_alignments();

    my @fragment_info;
    my $dead = 0;
    for (my $i=0; $i < @old_fragments; $i++) {
	unless ($new_fragment_hash{$old_fragments[$i]->get_id()} && (!$old_alignments[$i] || $alignment_mapping{$old_alignments[$i]->get_id()})) {
	    $dead = 1;
	    last;
	}
	push(@fragment_info, [$new_fragment_hash{$old_fragments[$i]->get_id()}, $old_alignments[$i] ?  $alignment_mapping{$old_alignments[$i]->get_id()} : undef]);
    } 

    if ($dead) {
	CLASSES::Pseudogene->mark_dead($dbh, $old_pgene->get_acc());
	print COMMENT_FILE "Pseudogene ".$old_pgene->get_id()." not remapped\n";
    }
    else {
	my $new_pgene = $old_pgene->new();
	$new_pgene->set_fragment_info(@fragment_info);
 	$new_pgene->insert();
    }
}
 
system("remap/remap-sets.pl $tax_id $source_build_number $target_build_number");
print("Done\n");




#######################
# Returns true if the fragments are on the same chromosome, at the same coordinates, and on the same strand.
sub compare_fragments {
    return 1 if $_[0]==$_[1];
    return 0 if $_[0]->get_chromosome() != $_[1]->get_chromosome() || $_[0]->get_strand() ne $_[1]->get_strand();
    my @c1 = $_[0]->get_coords();
    my @c2 = $_[1]->get_coords();
    return $c1[0]==$c2[0] && $c1[1]==$c2[1];
}


