package CLASSES::Protein;
use strict;
use CLASSES::DBClass_Versioned;
use QUERY::query_factory;
use TOOLS::assert;

our @VERSION = '1.0';
our @ISA = qw(CLASSES::DBClass_Versioned);

my $protein_id_table = TOOLS::constants::PROTEIN_ID_TABLE;

#-------------------------------------------------------------------------
# Required Functions
sub _vinit {
    my $self = shift;

    return $self;
}

#-------------------------------------------------------------------------
# Other Functions

#######################
# Name: get_primary_acc
# Purpose: (static) Find the "primary" acc for a given acc
# Parameters: Database handle, an acc
# Output: One acc
sub get_primary_acc {
   my $class = assert_static(shift);
   my $dbh = assert_dbh(shift);
   my $acc = shift;

   my $query = QUERY::Cache->get($dbh, $class, "get_primary_acc");
   unless ($query) {
       $query = create_query("type"=>"SVSelect",
	                     "dbh"=>$dbh,
			     "table"=>$protein_id_table,
                             "return_field"=>"_acc",
                             "where"=>"_secondary = ?_secondary?");
       QUERY::Cache->set($dbh, $class, "get_primary_acc", $query);
   }
   $query->execute("_secondary"=>$acc);
   return ($query->fetch_values())[0];
}

#######################
# Name: _is_primary_acc
# Purpose: (static) Determine if a given acc is primary
# Parameters: A database handle, an acc
# Output: A boolean
sub _is_primary_acc {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $acc = shift;

    my $query = QUERY::Cache->get($dbh, $class, "is_primary_acc");
    unless ($query) {
	$query = create_query("type"=>"SVSelect", 
			      "dbh"=>$dbh, 
			      "table"=>$protein_id_table,
			      "return_field"=>"_acc", 
			      "where"=>"_acc = ?where_acc?");
	QUERY::Cache->set($dbh, $class, "is_primary_acc", $query);
    }

    $query->execute("where_acc"=>$acc);
    return ($query->fetch_values() > 0);
}

#######################
# Name: _is_secondary_acc
# Purpose: (static) Determine if an acc is secondary
# Parameters: A database handle, an acc
# Output: Array of booleans in vector contact
sub _is_secondary_acc {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $acc = shift;

    my $primary = $class->get_primary_acc($dbh, $acc);
    return (defined($primary) && $primary ne $acc);
}

 ######################
# Name: add_secondaries
# Purpose: (static) Add new secondary/primary acc pair to the table
# Paramters: database handle, single primary acc, single secondary acc
# Output: None
sub add_secondaries {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $primary = shift;
    my $secondary = shift;

    my $query = QUERY::Cache->get($dbh, $class, "add_secondaries");
    unless ($query) {
	$query = create_query("type"=>"Insert", 
			      "dbh"=>$dbh, 
			      "table"=>$protein_id_table, 
			      "attributes"=>["_secondary","_acc"]);
	QUERY::Cache->set($dbh, $class, "add_secondaries", $query);
    }

    my $is_primary = $class->_is_primary_acc($dbh, $primary);
    my $is_secondary = $class->_is_secondary_acc($dbh, $secondary);

    if ($is_secondary) {
	assert(!$is_primary, "Cannot reassign a secondary acc to a new primary acc: $secondary\n");
    }
    else {
	$query->execute("_secondary"=>$secondary, "_acc"=>$primary);
    }
}

#######################
# Name: update_acc
# Purpose: Dynamic method to change the acc to the "primary" acc
# Parameters: None
# Output: The current acc (undef if acc is undefined)
sub update_acc {
    my $self = assert_dynamic(shift);

    my $old_acc = $self->get_acc();
    $self->set_acc(ref($self)->get_primary_acc($self->get_dbh(), $old_acc)) if defined($old_acc);
    return $self->get_acc();
}


#######################
# Name: insert
# Purpose: Augment CLASS:DBClass_Versioned::insert by updating the acc of all objects
# Parameters: No additional parameters
# Output: None
# Side effect: Each object will have its acc updated
sub insert {
    my $proto = shift;
    my ($class, $static) = ref($proto) ? (ref($proto), 0) : ($proto, 1);

    assert(@_==0, "Dynmic version takes no parameters") unless $static;
 
    my @objs = $static ? @_ : ($proto);
    foreach (@objs) {
	$_->update_acc();
    }
    $class->CLASSES::DBClass_Versioned::insert(@objs);
}

#######################
# Name: update
# Purpose: Augment CLASSES::DBClass_Versioned::update by updating the acc of the object
# Parameters: No additional parameters
# Output: None
# Side effect: Each object will have its acc updated
sub update {
    my $self = shift;
    $self->update_acc();
    $self->CLASSES::DBClass_Versioned::update();
}

#######################
# Name: remove
# Purpose: Augment CLASSES::DBClass_Versioned::remove by updating the acc of the object
# Parameters: No additional parameters
# Output: None
# Side effect: Each object will have its acc updated
sub remove {
    my $self = shift;
    $self->update_acc();
    $self->CLASSES::DBClass_Versioned::remove();
}

#######################
# Name: get_by_acc
# Purpose: Augment CLASS:DBClass_Versioned::get_by_acc by updating the given acc
# Parameters: No additional parameters
# Output: None
sub get_by_acc {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $acc = $class->get_primary_acc($dbh, shift);

    return $class->CLASSES::DBClass_Versioned::get_by_acc($dbh, $acc, "saved_query" => "protein_get_by_acc");
}


######################
# Name: make_new_primary
# Purpose: (static) Remap an acc group to a new primary
# Parameters: Database handle, one member of the acc group, the new primary, the id of the "latest" protein (optional)
# Output: None
sub make_new_primary {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $secondary_acc = shift;
    my $primary_acc = shift;
    my $final_id = shift;

    my $old_primary = $class->get_primary_acc($dbh, $secondary_acc);
    my $new_primary_map = $class->get_primary_acc($dbh, $primary_acc);
    #assert(!$new_primary_map || $new_primary_map ne $old_primary, "Cannot join acc groups: S:$secondary_acc P:$primary_acc OP:$old_primary NP:$new_primary_map");

    my $query = QUERY::Cache->get($dbh, $class, "make_new_primary");
    unless ($query) {
	$query = create_query("type"=>"Update",
			      "table"=>$protein_id_table,
			      "attributes"=>["_acc"],
			      "dbh"=>$dbh,
			      "where"=>"_acc=?old_acc?");
	QUERY::Cache->set($dbh, $class, "make_new_primary", $query);
    }
    $query->execute("old_acc"=>$old_primary, "_acc"=>$primary_acc);

    $class->add_secondaries($dbh, $primary_acc, $primary_acc) unless $new_primary_map;
    $class->_update_acc_to_primary($dbh, $old_primary, $final_id);
}

######################
# Name: join_acc_groups
# Purpose: (static) Join two or more acc groups (making the primary of the first the overall primary id)
# Parameters: A database handle, two or more primary accs (each should be from a seperate group)
# Output: None
# POSSILBE ERROR:  When merging groups, there needs to be a version-unqiue protein resulting
sub join_acc_groups {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my @accs = @_;

    assert(@accs > 1, "Need to provide at least two accs");

    my $new_primary = $class->get_primary_acc($dbh, shift @accs);

    my $query = QUERY::Cache->get($dbh, $class, "make_new_primary");
    unless ($query) {
	$query = create_query("type"=>"Update",
			      "table"=>$protein_id_table,
			      "attributes"=>["_acc"],
			      "dbh"=>$dbh,
			      "where"=>"_acc=?old_acc?");
	QUERY::Cache->set($dbh, $class, "make_new_primary", $query);
    }

    for (@accs) {
	$query->execute("old_acc"=>$class->get_primary_acc($dbh, $_), "_acc"=>$new_primary);
	$class->_update_acc_to_primary($dbh, $_);
    }
}

####################
# Name: _update_acc_to_primary
# Purpose: (static) Given an acc, change all occurences in the Protein and Protein_Version tables to the primary
# Paramters: Database handle, an acc, the "most recent" protein (optional)
# Output: None
sub _update_acc_to_primary {
    my $class = assert_static(shift);
    my $dbh = assert_dbh(shift);
    my $secondary = shift;
    my $final_id = shift;
    $final_id = $final_id->get_id() if $final_id && ref($final_id) eq 'CLASSES::Protein';
    
    assert(defined $secondary, "Need to provide an acc");
    assert($class->_is_secondary_acc($dbh, $secondary), "Need to provide an existing acc");
    
    my $primary = $class->get_primary_acc($dbh, $secondary);

    my @P;
    my $update_query_name = "_update_acc_query";
    @P = $class->get("dbh"=>$dbh, "saved_query" => $update_query_name, "where" => "acc=?acc1? or acc=?acc2?", "values" => {"acc1"=>$primary, "acc2"=>$secondary}, "all_versions" => 1);

    @P = sort {$a->get_version() <=> $b->get_version() || $a->get_entry_date() cmp $b->get_entry_date()} @P;
    if (@P > 1) {
       if ($final_id) {
          my $i=0;
          $i++ while ($i < @P && $P[$i]->get_id() != $final_id);
          assert($i < @P, "Final id not found");
          my $temp = $P[$i];
          while ($i < @P-1) {
             $P[$i] = $P[$i+1];
             $i++;
          }
          $P[-1] = $temp;
       }


	for (my $i=@P-1; $i>=0; $i--) {
	    my $v = $P[$i]->get_version();
	    if ($v != $i+1) {
		$P[$i]->_modify_version($i+1);
	    }
	}
    }

    if ($secondary ne $primary) {
    
	for my $P (grep {$_->get_acc() eq $secondary} @P) {
	    $P->_modify_object_in_db("acc", "$primary");
	}

        my $query_name = "remove_protein_version_acc";
	my $query = QUERY::Cache->get($dbh, $class, $query_name);
	unless ($query) {
	    $query = create_query(type => "Remove",
				  table => "Protein_Version",
				  "dbh" => $dbh,
				  "where" => "_acc=?sec?");
	    QUERY::Cache->set($dbh, $class, $query_name, $query);
	}
	$query->execute(sec=>$secondary);
 	$P[-1]->_update_current_version();
    }
}

######################
# Name: _merge_protein_accs
# Purpose: Given two independent protein accs, merge them to one (using the first as the new primary)
# Parameters: dbh, primary acc, other acc
# Output: None
sub _merge_protein_accs {
   my $class = assert_static(shift);
   my $dbh = assert_dbh(shift);
   assert(@_==2, "Requiers two accs");
   my ($primary, $secondary) = @_;
   assert($primary ne $secondary, "Acc parameters must be different");
   assert(CLASSES::Protein->get_primary_acc($dbh, $primary) eq $primary, "$primary (primary) not primary");
   assert(CLASSES::Protein->get_primary_acc($dbh, $secondary) eq $secondary, "$secondary (secondary) not primary");

   CLASSES::Protein->make_new_primary($dbh, $secondary, $primary, CLASSES::Protein->get_by_acc($dbh, $primary)); 
}

######################
# Name: get_acc_type
# Purpose: (dynamic) Give the type of accession number (Uniprot or Ensembl)
# Parameters: None
# Return: String containing the name
sub get_acc_type {
    my $self = assert_dynamic(shift);
    my $acc = $self->get_acc();

    
    return "Uniprot" if ($acc && $acc =~ /^\w\d\w[\w\d]{2}\d$/);
    return "Ensembl" if ($self->get_tax_id()==9606 && $acc && $acc =~ /^ENSP\d{11}/);
    return "";
}

######################
# Name: get_by_name

sub get_by_name {
	my $class = assert_static(shift);
	my $dbh = assert_dbh(shift);
	my $name = shift;

	my @where = ("name=?name_id?");
	return $class->get(dbh => $dbh, where => \@where, values => {name_id => $name});
}

######################
# Name: get_link
# Purpose: (dynamic) Return an html link to the generating database (if known)
# Parameters: None
# Ouput: An html link (or empty string if none known)
sub get_link {
    my $self = assert_dynamic(shift);
    my $acc = $self->get_acc();
    my $type = $self->get_acc_type();

    return "" unless $acc;

    if ($type eq "Uniprot") {
	return "http://www.uniprot.org/entry/$acc";
    }

    if ($type eq "Ensembl" && $self->get_tax_id()==9606) {
	return "http://www.ensembl.org/Homo_sapiens/protview?db=core;peptide=$acc";
    }

    return "";
}

1;
